from __future__ import annotations

from typing import Union, Any

from spinta.backends.constants import BackendFeatures
from spinta.components import Property, Page, UrlParams
from spinta.core.ufuncs import Expr, asttoexpr
from spinta.datasets.components import ExternalBackend
from spinta.types.datatype import Denorm, Ref, BackRef
from spinta.types.text.components import Text
from spinta.types.text.helpers import determine_language_property_for_text
from spinta.ufuncs.components import ForeignProperty
from spinta.ufuncs.querybuilder.components import QueryBuilder, QueryParams, QueryPage, LiteralProperty, Star
from spinta.ufuncs.helpers import merge_formulas
from spinta.utils.types import is_value_literal


def expanded(env: QueryBuilder, prop: Property):
    # If backend does not support expand, assume it is always expanded
    if not env.backend.supports(BackendFeatures.EXPAND):
        return True

    # If property does not support expandability, assume it's expanded by default
    if not prop.dtype.expandable:
        return True

    # None - nothing is expanded
    if env.expand is None:
        return False

    # Star - everything is expanded
    if isinstance(env.expand, Star):
        return True

    return prop in env.expand


def get_language_column(env: QueryBuilder, dtype: Text):
    default_langs = env.context.get('config').languages
    prop = determine_language_property_for_text(dtype, env.query_params.lang_priority, default_langs)
    column = env.backend.get_column(env.table, prop)
    return column


def get_column_with_extra(env: QueryBuilder, prop: Property):
    if isinstance(prop.dtype, Text):
        return get_language_column(env, prop.dtype)
    return env.backend.get_column(env.table, prop)


def get_page_values(env: QueryBuilder, row: dict):
    if not env.page.page_.filter_only:
        if isinstance(env.model.backend, ExternalBackend):
            return [row[item.prop.external.name] for item in env.page.page_.by.values()]
        else:
            return [row[item.prop.name] for item in env.page.page_.by.values()]


def merge_with_page_selected_list(select_list: list, page: QueryPage):
    merged_selected = select_list or []
    if page.page_.enabled and page.select is not None:
        for select in page.select:
            if select not in merged_selected:
                merged_selected.append(select)
    return merged_selected


def merge_with_page_sort(sort: list, page: QueryPage):
    merged_sort = sort or []
    if page.page_.enabled and page.sort:
        merged_sort = page.sort
    return merged_sort


def merge_with_page_limit(limit: int, page: QueryPage):
    if page and page.size and page.page_.enabled:
        if limit and limit < page.size:
            return limit
        return page.size
    return limit


def update_query_with_url_params(query_params: QueryParams, url_params: UrlParams):
    query_params.prioritize_uri = url_params.fmt.prioritize_uri
    query_params.lang_priority = url_params.accept_langs
    query_params.lang = url_params.lang
    query_params.expand = url_params.expand


def _create_or_condition(condition_info: list):
    action = condition_info[0]
    page_by = condition_info[1]
    or_null = condition_info[2]

    result = {
        'name': action,
        'args': [{
            'name': 'bind',
            'args': [page_by.prop.name]
        }, page_by.value]
    }

    if or_null != -1:
        result = {
            'name': 'or',
            'args': [
                result,
                {
                    'name': 'eq' if or_null == 1 else 'ne',
                    'args': [{
                        'name': 'bind',
                        'args': [page_by.prop.name]
                    }, None]
                }
            ]}
    return result


def _get_null_action(by: str, needs_null: bool = False):
    if not needs_null:
        return -1

    if by.startswith('-'):
        return 0
    return 1


def get_pagination_compare_query(
    model_page: Page,
) -> Union[Expr, None]:
    filtered = model_page
    item_count = len(filtered.by.keys())
    where_list = []
    for i in range(item_count):
        where_list.append([])
    if not filtered.all_none() or not filtered.first_time:
        for i, (by, page_by) in enumerate(filtered.by.items()):
            for n in range(item_count):
                if n >= i:
                    if n == i:
                        if page_by.value is not None:
                            if by.startswith('-'):
                                where_list[n].append(('lt', page_by, _get_null_action(by)))
                            else:
                                where_list[n].append(('gt', page_by, _get_null_action(by, True)))
                        else:
                            if by.startswith('-'):
                                where_list[n].append(('ne', page_by, _get_null_action(by)))
                    else:
                        where_list[n].append(('eq', page_by, _get_null_action(by)))
    remove_list = []

    for i, (by, value) in enumerate(filtered.by.items()):
        if value.value is None and not by.startswith('-'):
            remove_list.append(where_list[i])
    for item in remove_list:
        where_list.remove(item)

    where_compare = {}
    for where in where_list:
        compare = {}
        for item in where:
            condition = _create_or_condition(item)
            if compare:
                compare = {
                    'name': 'and',
                    'args': [
                        compare,
                        condition
                    ]
                }
            else:
                compare = condition
        if where_compare:
            where_compare = {
                'name': 'or',
                'args': [
                    where_compare,
                    compare
                ]
            }
        else:
            where_compare = compare

    if where_compare:
        where_compare = asttoexpr(where_compare)

    query = where_compare or None
    return query


def add_page_expr(expr: Expr, page: Page):
    return merge_formulas(expr, asttoexpr({
        'name': 'paginate',
        'args': [
            page
        ]
    }))


def process_literal_value(value: Any) -> Any:
    if is_value_literal(value):
        return LiteralProperty(value)
    return value


def denorm_to_foreign_property(dtype: Denorm) -> ForeignProperty | Property:
    ref = dtype.prop.parent
    root_parent = ref
    if isinstance(ref, Property) and isinstance(ref.dtype, (Ref, BackRef)):
        fpr = None
        if ref.dtype.inherited:
            parent_list = []
            root_ref_parent = ref
            while root_ref_parent and isinstance(root_ref_parent, Property) and isinstance(root_ref_parent.dtype, (Ref, BackRef)):
                parent_list.append(root_ref_parent)
                if not root_ref_parent.dtype.inherited:
                    break
                root_ref_parent = root_ref_parent.parent

            if parent_list:
                parent_list = list(reversed(parent_list))
                root_parent = parent_list.pop(0)
                for parent in parent_list:
                    fpr = ForeignProperty(fpr, root_parent.dtype, parent.dtype)
                    root_parent = parent

        return ForeignProperty(fpr, root_parent.dtype, dtype.rel_prop.dtype)
    return dtype.rel_prop



import io
import pathlib
from typing import Any
from typing import Dict
from typing import Iterator

from starlette.requests import Request
from starlette.responses import Response
from starlette.responses import StreamingResponse

from spinta import commands
from spinta.components import Action
from spinta.components import Context
from spinta.components import Model
from spinta.components import UrlParams
from spinta.formats.helpers import get_model_tabular_header
from spinta.formats.xlsx.components import Xlsx
from spinta.manifests.components import Manifest
from spinta.manifests.tabular.components import ManifestRow
from spinta.manifests.tabular.constants import DATASET
from spinta.manifests.tabular.helpers import datasets_to_tabular, write_xlsx
from spinta.utils.nestedstruct import flatten, sepgetter
from spinta.utils.response import aiter


@commands.render.register(Context, Request, Model, Xlsx)
def render(
    context: Context,
    request: Request,
    model: Model,
    fmt: Xlsx,
    *,
    action: Action,
    params: UrlParams,
    data: Iterator[Dict[str, Any]],
    status_code: int = 200,
    headers: Dict[str, str] = None,
) -> Response:
    headers = headers or {}
    headers['Content-Disposition'] = f'attachment; filename="{model.basename}.xlsx"'

    rows = flatten(data, sepgetter(model))
    cols = get_model_tabular_header(context, model, action, params)
    return StreamingResponse(
        aiter(_render_xlsx(rows, cols)),
        status_code=status_code,
        media_type=fmt.content_type,
        headers=headers,
    )


@commands.render.register(Context, Manifest, Xlsx)
def render(
    context: Context,
    manifest: Manifest,
    fmt: Xlsx,
    *,
    action: Action = None,
    params: UrlParams = None,
    status_code: int = 200,
    headers: Dict[str, str] = None,
    path: str = None
) -> Response:
    rows = datasets_to_tabular(context, manifest)
    rows = ({c: row[c] for c in DATASET} for row in rows)
    if not path:
        headers = headers or {}
        headers['Content-Disposition'] = f'attachment; filename="{manifest.name}.xlsx"'
        return StreamingResponse(
            aiter(_render_xlsx(rows, DATASET)),
            status_code=status_code,
            media_type=fmt.content_type,
            headers=headers,
        )
    else:
        write_xlsx(pathlib.Path(path), rows, DATASET)


def _render_xlsx(
    rows: Iterator[ManifestRow],
    cols: list
):
    stream = io.BytesIO()
    write_xlsx(stream, rows, cols)
    stream.seek(0)
    for row in stream.readlines():
        yield row



from spinta import commands
from spinta.components import Context
from spinta.manifests.components import Manifest


@commands.load_for_request.register(Context, Manifest)
def load_for_request(context: Context, manifest: Manifest):
    pass


@commands.fully_initialize_manifest.register(Context, Manifest)
def fully_initialize_manifest(context: Context, manifest: Manifest):
    pass


@commands.create_request_manifest.register(Context, Manifest)
def create_request_manifest(context: Context, manifest: Manifest):
    return manifest



from abc import ABC, abstractmethod
from typing import Dict, List
import sqlalchemy as sa
from spinta import commands
from spinta.components import Model, Namespace, Context
from spinta.datasets.components import Dataset
from spinta.datasets.inspect.helpers import zipitems
from spinta.exceptions import ModelNotFound, NamespaceNotFound, DatasetNotFound
from spinta.manifests.components import Manifest
from spinta.manifests.internal_sql.components import InternalSQLManifest
from spinta.manifests.internal_sql.helpers import internal_to_schema, load_internal_manifest_nodes, get_object_from_id, \
    select_full_table, update_schema_with_external, load_required_models, get_manifest, get_transaction_connection, \
    datasets_to_sql
from spinta.types.namespace import load_namespace_from_name
from spinta.utils.schema import NA


def get_model_name_list(context: Context, manifest: InternalSQLManifest, loaded: bool, namespace: str = None,
                        recursive: bool = False):
    manifest = get_manifest(context, manifest)
    table = manifest.table
    conn = get_transaction_connection(context)
    objs = manifest.get_objects()
    if conn is None or loaded:
        models = objs.get('model', None)
        if models is None:
            return

        if namespace:
            for model_name, model in objs['model'].items():
                if model.ns.name == namespace:
                    yield model_name
        else:
            yield from objs['model'].keys()

        # Escape generator
        return

    # Include meta models
    if namespace == '':
        for model in objs['model'].values():
            if model.name.startswith('_'):
                yield model.name

    # Check namespace if given, else do normal
    stmt = sa.select(table.c.path).where(
        sa.and_(
            table.c.path.startswith(namespace),
            table.c.dim == 'model'
        )
        if namespace
        else table.c.dim == 'model'
    ).order_by(table.c.index)

    rows = conn.execute(stmt)
    for row in rows:
        if not recursive and namespace:
            # Check if path is actually right after ns,
            # ex: namespace = 'dataset/test'
            # models: 'dataset/test/gov/Model', 'dataset/test/Model'
            # This will filter out first model, since it belongs to gov namespace
            fixed_path = row['path'].replace(namespace, '')
            if fixed_path.startswith('/'):
                fixed_path = fixed_path[1:]
            if len(fixed_path.split('/')) == 1:
                yield row['path']
        else:
            yield row['path']


def get_namespace_name_list(context: Context, manifest: InternalSQLManifest, loaded: bool, namespace: str = None):
    manifest = get_manifest(context, manifest)
    table = manifest.table
    conn = get_transaction_connection(context)
    if conn is None or loaded:
        objs = manifest.get_objects()
        namespaces = objs.get('ns', None)
        if namespaces is None:
            return

        if namespace is not None:
            for ns_name, ns in namespaces.items():
                if ns.parent and isinstance(ns.parent, Namespace) and ns.parent.name == namespace:
                    yield ns_name
        else:
            yield from namespaces.keys()

        # Escape generator
        return

    # If namespace given, select all namespace/datasets that starts with, but it's not it.
    stmt = sa.select(table.c.mpath).where(
        sa.and_(
            table.c.mpath.startswith(namespace),
            table.c.mpath != namespace,
            sa.or_(
                table.c.dim == 'namespace',
                table.c.dim == 'dataset'
            )
        )
        if namespace
        else sa.or_(
            table.c.dim == 'namespace',
            table.c.dim == 'dataset'
        )
    ).order_by(table.c.mpath)

    rows = conn.execute(stmt)
    yielded = []
    for row in rows:
        if namespace:
            # Fix namespace path, ex given namespace is 'dataset/test'
            # Fetched namespaces are: 'dataset/test/gov', 'dataset/test/other/gov'
            # it will return 'dataset/test/gov' and 'dataset/test/other'
            fixed_path = row['mpath'].replace(namespace, '')
            if fixed_path.startswith('/'):
                fixed_path = fixed_path[1:]
            fixed_path = f'{namespace}/{fixed_path.split("/")[0]}'
            if fixed_path not in yielded:
                yielded.append(fixed_path)
                yield fixed_path
        else:
            yield row['mpath']


def _get_dataset_name_list(context: Context, manifest: InternalSQLManifest, loaded: bool):
    manifest = get_manifest(context, manifest)
    table = manifest.table
    conn = get_transaction_connection(context)
    if conn is None or loaded:
        objs = manifest.get_objects()
        if 'dataset' and objs and objs['dataset']:
            yield from objs['dataset'].keys()

        # Exit generator
        return

    stmt = sa.select(table.c.path).where(
        table.c.dim == 'dataset'
    ).order_by(table.c.path)
    rows = conn.execute(stmt)
    for row in rows:
        yield row['path']


@commands.has_model.register(Context, InternalSQLManifest, str)
def has_model(context: Context, manifest: InternalSQLManifest, model: str, loaded: bool = False, **kwargs):
    manifest = get_manifest(context, manifest)
    if model in manifest.get_objects()['model']:
        return True

    conn = get_transaction_connection(context)
    if not loaded and conn is not None:
        table = manifest.table
        return conn.execute(sa.select([
            sa.exists().where(
                sa.and_(
                    table.c.path == model,
                    table.c.dim == 'model'
                )
            )
        ])).scalar()
    return False


@commands.get_model.register(Context, InternalSQLManifest, str)
def get_model(context: Context, manifest: InternalSQLManifest, model: str, **kwargs):
    manifest = get_manifest(context, manifest)

    # Check if model exists
    if not has_model(context, manifest, model):
        raise ModelNotFound(model=model)

    conn = get_transaction_connection(context)
    objects = manifest.get_objects()

    if model in objects['model']:
        return objects['model'][model]

    if conn is not None:
        schemas = []
        table = manifest.table
        model_obj = conn.execute(
            select_full_table(table).where(
                sa.and_(
                    table.c.path == model,
                    table.c.dim == 'model'
                )
            )
        ).one()
        props = conn.execute(
            select_full_table(table).where(
                sa.and_(
                    table.c.path.startswith(f'{model}/'),
                    table.c.dim != 'model'
                )
            ).order_by(table.c.index)
        )

        parent_id = model_obj['parent']
        parent_dataset = None
        parent_resource = None
        parent_schemas = []
        while parent_id is not None:
            parent_obj = get_object_from_id(context, manifest, parent_id)
            if parent_obj is None:
                break

            parent_schemas.append(parent_obj)
            parent_id = parent_obj['parent']

            if parent_obj['dim'] == 'dataset':
                parent_dataset = parent_obj['name']
                dataset = commands.get_dataset(context, manifest, parent_dataset)
                if parent_resource:
                    get_dataset_resource(context, manifest, dataset, parent_resource)
                break
            elif parent_obj['dim'] == 'resource' and not parent_resource:
                parent_resource = parent_obj['name']

        schemas.extend(reversed(parent_schemas))
        schemas.append(model_obj)
        schemas.extend(props)
        required_models = [model]

        schemas = internal_to_schema(manifest, schemas)
        schemas = update_schema_with_external(schemas, {
            'dataset': parent_dataset,
            'resource': parent_resource
        })
        schemas = load_required_models(context, manifest, schemas, required_models)
        load_internal_manifest_nodes(context, manifest, schemas, link=True, ignore_types=['dataset', 'resource'])
        if model in objects['model']:
            new_model = objects['model'][model]
            commands.prepare(context, manifest.backend, new_model, ignore_duplicate=True)
            return new_model


@commands.get_models.register(Context, InternalSQLManifest)
def get_models(context: Context, manifest: InternalSQLManifest, loaded: bool = False, **kwargs):
    model_names = get_model_name_list(context, manifest, loaded)
    objs = manifest.get_objects()
    for name in model_names:
        # get_model loads the model if it has not been loaded
        if name not in objs['model']:
            commands.get_model(context, manifest, name)
    return objs['model']


@commands.set_model.register(Context, InternalSQLManifest, str, Model)
def set_model(context: Context, manifest: InternalSQLManifest, model_name: str, model: Model, **kwargs):
    manifest.get_objects()['model'][model_name] = model


@commands.set_models.register(Context, InternalSQLManifest, dict)
def set_models(context: Context, manifest: InternalSQLManifest, models: Dict[str, Model], **kwargs):
    manifest.get_objects()['model'] = models


@commands.has_namespace.register(Context, InternalSQLManifest, str)
def has_namespace(context: Context, manifest: InternalSQLManifest, namespace: str, loaded: bool = False, **kwargs):
    manifest = get_manifest(context, manifest)
    conn = get_transaction_connection(context)

    if namespace in manifest.get_objects()['ns']:
        return True

    if conn is not None and not loaded:
        table = manifest.table
        return conn.execute(sa.select([
            sa.exists().where(
                sa.or_(
                    table.c.mpath == namespace,
                    table.c.mpath.startswith(f'{namespace}/')
                )
            )
        ])).scalar()
    return False


@commands.get_namespace.register(Context, InternalSQLManifest, str)
def get_namespace(context: Context, manifest: InternalSQLManifest, namespace: str, **kwargs):
    manifest = get_manifest(context, manifest)

    # Check if namespace exists
    if not has_namespace(context, manifest, namespace):
        raise NamespaceNotFound(namespace=namespace)

    conn = get_transaction_connection(context)
    objects = manifest.get_objects()

    if namespace in objects['ns']:
        return objects['ns'][namespace]

    if conn is not None:
        table = manifest.table
        ns = conn.execute(
            select_full_table(table).where(
                sa.and_(
                    table.c.name == namespace,
                    table.c.dim == 'ns'
                )
            ).order_by(table.c.index)
        )
        schemas = internal_to_schema(manifest, ns)
        load_internal_manifest_nodes(context, manifest, schemas, link=True)
        if namespace in objects['ns']:
            return objects['ns'][namespace]

        return load_namespace_from_name(context, manifest, namespace, drop=False)


@commands.get_namespaces.register(Context, InternalSQLManifest)
def get_namespaces(context: Context, manifest: InternalSQLManifest, loaded: bool = False, **kwargs):
    ns_names = get_namespace_name_list(context, manifest, loaded)
    objs = manifest.get_objects()
    for name in ns_names:
        # get_namespace loads the namespace if it has not been loaded
        if name not in objs['ns']:
            commands.get_namespace(context, manifest, name)
    return objs['ns']


@commands.set_namespace.register(Context, InternalSQLManifest, str, Namespace)
def set_namespace(context: Context, manifest: InternalSQLManifest, namespace: str, ns: Namespace, **kwargs):
    manifest = get_manifest(context, manifest)
    manifest.get_objects()['ns'][namespace] = ns


@commands.has_dataset.register(Context, InternalSQLManifest, str)
def has_dataset(context: Context, manifest: InternalSQLManifest, dataset: str, loaded: bool = False, **kwargs):
    manifest = get_manifest(context, manifest)
    conn = get_transaction_connection(context)
    if dataset in manifest.get_objects()['dataset']:
        return True

    if conn is not None and not loaded:
        table = manifest.table
        return conn.execute(sa.select([
            sa.exists().where(
                sa.and_(
                    table.c.mpath == dataset,
                    table.c.dim == 'dataset'
                )
            )
        ])).scalar()
    return False


def has_dataset_resource(context: Context, manifest: InternalSQLManifest, dataset: Dataset, resource: str, **kwargs):
    manifest = get_manifest(context, manifest)
    conn = get_transaction_connection(context)
    if resource in dataset.resources:
        return True

    if conn is not None:
        table = manifest.table
        return conn.execute(sa.select([
            sa.exists().where(
                sa.and_(
                    table.c.path == dataset.name,
                    table.c.dim == 'resource',
                    table.c.name == resource
                )
            )
        ])).scalar()
    return False


def get_dataset_resource(context: Context, manifest: InternalSQLManifest, dataset: Dataset, resource: str, **kwargs):
    manifest = get_manifest(context, manifest)

    # Check if dataset has resource
    if not has_dataset_resource(context, manifest, dataset, resource, **kwargs):
        return

    if resource in dataset.resources:
        return dataset.resources[resource]

    conn = get_transaction_connection(context)
    if conn is not None:
        table = manifest.table
        resources = conn.execute(
            sa.select(table).where(
                sa.and_(
                    table.c.path == dataset.name,
                    sa.or_(
                        sa.and_(
                            table.c.name == resource,
                            table.c.dim == 'resource'
                        ),
                        table.c.dim == 'dataset'
                    )
                )
            ).order_by(table.c.index)
        )
        schemas = internal_to_schema(manifest, resources)
        load_internal_manifest_nodes(context, manifest, schemas, link=True)
        if resource in dataset.resources:
            return dataset.resources[resource]


@commands.get_dataset.register(Context, InternalSQLManifest, str)
def get_dataset(context: Context, manifest: InternalSQLManifest, dataset: str, **kwargs):
    manifest = get_manifest(context, manifest)

    # Check if dataset exists
    if not has_dataset(context, manifest, dataset):
        raise DatasetNotFound(dataset=dataset)

    conn = get_transaction_connection(context)
    objects = manifest.get_objects()

    if dataset in objects['dataset']:
        return objects['dataset'][dataset]

    if conn is not None:
        table = manifest.table
        ds = conn.execute(
            select_full_table(table).where(
                sa.and_(
                    table.c.path == dataset,
                    table.c.dim != 'base',
                    table.c.dim != 'resource',
                )
            ).order_by(table.c.index)
        )
        schemas = internal_to_schema(manifest, ds)
        load_internal_manifest_nodes(context, manifest, schemas, link=True)
        if dataset in objects['dataset']:
            return objects['dataset'][dataset]


@commands.get_datasets.register(Context, InternalSQLManifest)
def get_datasets(context: Context, manifest: InternalSQLManifest, loaded: bool = False, **kwargs):
    dataset_names = _get_dataset_name_list(context, manifest, loaded)
    objs = manifest.get_objects()
    for name in dataset_names:
        # get_dataset loads the dataset if it has not been loaded
        if name not in objs['dataset']:
            commands.get_dataset(context, manifest, name)
    return objs['dataset']


@commands.set_dataset.register(Context, InternalSQLManifest, str, Dataset)
def set_dataset(context: Context, manifest: InternalSQLManifest, dataset_name: str, dataset: Dataset, **kwargs):
    manifest.get_objects()['dataset'][dataset_name] = dataset


@commands.get_dataset_models.register(Context, InternalSQLManifest, str)
def get_dataset_models(context: Context, manifest: InternalSQLManifest, dataset_name: str, **kwargs):
    model_names = get_model_name_list(context, manifest, False, dataset_name)
    if commands.has_dataset(context, manifest, dataset_name):
        dataset = commands.get_dataset(context, manifest, dataset_name)
        filtered_list = {}
        for name in model_names:
            model = commands.get_model(context, manifest, name)
            if model.external and model.external.dataset == dataset:
                filtered_list[name] = model
        return filtered_list
    return {}


class RowMergeAction(ABC):
    row: dict

    def __init__(self, row):
        self.row = row

    @abstractmethod
    def query(self, table):
        pass

    def update_index(self, index):
        self.row['index'] = index
        return index + 1


class RowMergeInsertAction(RowMergeAction):
    def query(self, table):
        return sa.insert(table).values(self.row)


class RowMergeUpdateAction(RowMergeAction):
    def query(self, table):
        return sa.update(table).where(table.c.id == self.row['id']).values(self.row)


class RowMergeDeleteAction(RowMergeAction):
    def query(self, table):
        return sa.delete(table).where(table.c.id == self.row['id'])

    def update_index(self, index):
        return index


def _adjust_indexes(table, conn, from_index, starting_index):
    if from_index is not None:
        rows = conn.execute(
            select_full_table(table).where(
                table.c.index > from_index
            )
        )
        index = starting_index + 1
        for row in rows:
            conn.execute(
                sa.update(table).where(table.c.id == row['id']).values(
                    index=index
                )
            )
            index += 1


class RowMergeBuilder:
    actions: List[RowMergeAction]

    def __init__(self):
        self.actions = []

    def add_action(self, action: RowMergeAction):
        self.actions.append(action)

    def execute_all(self, table, conn, starting_index=None, last_index=None):
        index = starting_index
        if index is None:
            index = 0

        new_last_index = None
        for action in self.actions:
            index = action.update_index(index)
            new_last_index = index

        if last_index is not None and new_last_index is not None:
            _adjust_indexes(table, conn, last_index, new_last_index)

        for action in self.actions:
            conn.execute(
                action.query(table)
            )


@commands.update_manifest_dataset_schema.register(Context, InternalSQLManifest, Manifest)
def update_manifest_dataset_schema(context: Context, manifest: InternalSQLManifest, target_manifest: Manifest,
                                   **kwargs):
    manifest = get_manifest(context, manifest)
    conn = get_transaction_connection(context)

    datasets = commands.get_datasets(context, target_manifest)
    if conn is None:
        raise Exception("NO CONNECTION")

    table = manifest.table
    target_rows = datasets_to_sql(context, target_manifest)

    for dataset in datasets.values():
        in_memory_existing_rows = []
        initial_index = None
        last_index = None
        if has_dataset(context, manifest, dataset.name):
            existing_data = conn.execute(
                select_full_table(table).where(
                    table.c.path.startswith(dataset.name),
                ).order_by(table.c.index)
            )

            for row in existing_data:
                in_memory_existing_rows.append(row)
                if initial_index is None:
                    initial_index = row['index']
                last_index = row['index']

        else:
            last_index_row = conn.execute(
                select_full_table(table)
                .order_by(sa.desc(table.c.index)).limit(1)
            )
            initial_index = None
            for row in last_index_row:
                initial_index = row['index'] + 1

        zipped_rows = zipitems(
            target_rows,
            in_memory_existing_rows,
            _id_key
        )
        merge_builder = RowMergeBuilder()
        models_to_unload = []
        for rows in zipped_rows:
            for new, old in rows:
                if old is NA and new is not NA:
                    # Insert action
                    merge_builder.add_action(
                        RowMergeInsertAction(new)
                    )
                elif old is not NA and new is not NA:
                    # Update action
                    merge_builder.add_action(
                        RowMergeUpdateAction(new)
                    )
                else:
                    # Delete action
                    merge_builder.add_action(
                        RowMergeDeleteAction(old)
                    )
                    if old['dim'] == 'model':
                        if old['path'] not in models_to_unload:
                            models_to_unload.append(old['path'])

        with manifest.engine.begin() as merge_connection:
            merge_builder.execute_all(
                table,
                merge_connection,
                initial_index,
                last_index
            )


def _id_key(row: dict) -> str:
    return row["id"]



from __future__ import annotations

from spinta.datasets.backends.sql.backends.sqlite.components import Sqlite
from spinta.datasets.backends.sql.ufuncs.query.components import SqlQueryBuilder


class SqliteQueryBuilder(SqlQueryBuilder):
    backend: Sqlite



from __future__ import annotations

from typing import List, Any, Tuple, Dict

from spinta.components import Page, Property
from spinta.core.ufuncs import ufunc, Expr, Negative, Bind, GetAttr
from spinta.datasets.backends.sql.ufuncs.components import Selected
from spinta.datasets.components import ExternalBackend
from spinta.exceptions import InvalidArgumentInExpression, CannotSelectTextAndSpecifiedLang
from spinta.types.datatype import DataType, String, PrimaryKey, Denorm
from spinta.types.text.components import Text
from spinta.ufuncs.components import ForeignProperty
from spinta.ufuncs.querybuilder.components import QueryBuilder, Star, ReservedProperty, NestedProperty, \
    ResultProperty, LiteralProperty, Flip, Func
from spinta.ufuncs.querybuilder.helpers import get_pagination_compare_query, process_literal_value, \
    denorm_to_foreign_property
from spinta.utils.schema import NA

# This file contains reusable resolvers, that should be backend independent
# in case there are cases where you need to have backend specific, just overload them
# keep in mind, that `select(env, expr)` is written backend specific, at least for now

# The main goal of this resolver logic, is that you have `_resolve_getattr`
# which is responsible for returning right types to process further
# Bind mean that it's a leaf node and can be accessed directly
# GetAttr mean that it can be nested further
# ForeignProperty is used, when you need to access properties from other models
# ReservedProperty is used, when datatypes do not have direct association to required field
#   example: File (when accessing child properties), Ref (when accessing it's own key, without needing to join models)


COMPARE = [
    'eq',
    'ne',
    'lt',
    'le',
    'gt',
    'ge',
    'startswith',
    'contains',
]


@ufunc.resolver(QueryBuilder, Denorm)
def _denorm_to_foreign_property(
    env: QueryBuilder,
    dtype: Denorm
):
    return denorm_to_foreign_property(dtype)


@ufunc.resolver(QueryBuilder, Bind, Bind, name='getattr')
def getattr_(
    env: QueryBuilder,
    field: Bind,
    attr: Bind
):
    return GetAttr(field.name, attr)


@ufunc.resolver(QueryBuilder, Bind, GetAttr, name='getattr')
def getattr_(
    env: QueryBuilder,
    obj: Bind,
    attr: GetAttr
):
    return GetAttr(obj.name, attr)


@ufunc.resolver(QueryBuilder, GetAttr, Bind, name='getattr')
def getattr_(
    env: QueryBuilder,
    obj: GetAttr,
    attr: Bind
):
    leaf = env.call('getattr', obj.name, attr)
    return GetAttr(obj.obj, leaf)


@ufunc.resolver(QueryBuilder, GetAttr)
def select(env: QueryBuilder, attr: GetAttr) -> Selected:
    resolved = env.resolve_property(attr)
    return env.call('select', resolved)


@ufunc.resolver(QueryBuilder, NestedProperty)
def select(env: QueryBuilder, nested: NestedProperty) -> Selected:
    return Selected(
        prop=nested.left.prop,
        prep=env.call('select', nested.right),
    )


@ufunc.resolver(QueryBuilder, ReservedProperty)
def select(env, prop):
    return env.call('select', prop.dtype, prop)


@ufunc.resolver(QueryBuilder, DataType, ReservedProperty)
def select(env, dtype, reserved):
    return env.call('select', dtype, reserved.param)


@ufunc.resolver(QueryBuilder, ResultProperty)
def select(env: QueryBuilder, prop: ResultProperty):
    return Selected(
        prep=prop.expr
    )


@ufunc.resolver(QueryBuilder, ForeignProperty)
def select(
    env: QueryBuilder,
    fpr: ForeignProperty,
) -> Selected:
    return env.call('select', fpr, fpr.right.prop)


@ufunc.resolver(QueryBuilder, ForeignProperty, Func)
def select(
    env: QueryBuilder,
    fpr: ForeignProperty,
    func: Func
) -> Selected:
    return env.call('select', fpr, fpr.right.prop, func)


@ufunc.resolver(QueryBuilder, ForeignProperty, Property, Func)
def select(
    env: QueryBuilder,
    fpr: ForeignProperty,
    prop: Property,
    func: Func
) -> Selected:
    return env.call('select', fpr, prop.dtype, func)


@ufunc.resolver(QueryBuilder, ForeignProperty, Property)
def select(
    env: QueryBuilder,
    fpr: ForeignProperty,
    prop: Property,
) -> Selected:
    resolved_key = fpr / prop
    if resolved_key not in env.resolved:
        if isinstance(prop.external, list):
            raise ValueError("Source can't be a list, use prepare instead.")
        if prop.external and prop.external.prepare is not NA:
            if isinstance(prop.external.prepare, Expr):
                result = env(this=prop).resolve(prop.external.prepare)
            else:
                result = prop.external.prepare
            result = env.call('select', fpr, prop.dtype, result)
        else:
            result = env.call('select', fpr, prop.dtype)
        assert isinstance(result, Selected), prop
        env.resolved[resolved_key] = result
    return env.resolved[resolved_key]


@ufunc.resolver(QueryBuilder, ForeignProperty, DataType, object)
def select(
    env: QueryBuilder,
    fpr: ForeignProperty,
    dtype: DataType,
    prep: Any,
) -> Selected:
    if isinstance(prep, str):
        # XXX: Backwards compatibility thing.
        #      str values are interpreted as Bind values and Bind values are
        #      assumed to be properties. So here we skip
        #      `env.call('select', prep)` and return `prep` as is.
        #      This should be eventually removed, once backwards compatibility
        #      for resolving strings as properties is removed.
        return Selected(prop=dtype.prop, prep=prep)
    elif isinstance(prep, Expr):
        # If `prepare` expression returns another expression, then this means,
        # it must be processed on values returned by query.
        prop = dtype.prop
        if not isinstance(env.backend, ExternalBackend) or (prop.external and prop.external.name) or prop.dtype.inherited:
            sel = env.call('select', fpr, dtype)
            return Selected(item=sel.item, prop=sel.prop, prep=prep)
        else:
            return Selected(item=None, prop=prop, prep=prep)
    else:
        result = env.call('select', fpr, prep)
        return Selected(prop=dtype.prop, prep=result)


@ufunc.resolver(QueryBuilder, ForeignProperty, tuple)
def select(
    env: QueryBuilder,
    fpr: ForeignProperty,
    prep: Tuple[Any],
) -> Tuple[Any]:
    return tuple(env.call('select', fpr, v) for v in prep)


@ufunc.resolver(QueryBuilder, ForeignProperty, PrimaryKey)
def select(
    env: QueryBuilder,
    fpr: ForeignProperty,
    dtype: PrimaryKey,
) -> Selected:
    model = dtype.prop.model
    pkeys = model.external.pkeys

    if not pkeys:
        raise RuntimeError(
            f"Can't join {dtype.prop} on right table without primary key."
        )

    if len(pkeys) == 1:
        prop = pkeys[0]
        result = env.call('select', fpr, prop)
    else:
        result = [
            env.call('select', fpr, prop)
            for prop in pkeys
        ]
    return Selected(prop=dtype.prop, prep=result)


@ufunc.resolver(QueryBuilder, list)
def select(
    env: QueryBuilder,
    prep: List[Any],
) -> List[Any]:
    return [env.call('select', v) for v in prep]


@ufunc.resolver(QueryBuilder, tuple)
def select(
    env: QueryBuilder,
    prep: Tuple[Any],
) -> Tuple[Any]:
    return tuple(env.call('select', v) for v in prep)


@ufunc.resolver(QueryBuilder, dict)
def select(
    env: QueryBuilder,
    prep: Dict[str, Any],
) -> Dict[str, Any]:
    # TODO: Add tests.
    return {k: env.call('select', v) for k, v in prep.items()}


@ufunc.resolver(QueryBuilder, LiteralProperty)
def select(
    env: QueryBuilder,
    prep: LiteralProperty,
) -> Selected:
    return Selected(prep=prep.value)


@ufunc.resolver(QueryBuilder, DataType, object)
def select(env: QueryBuilder, dtype: DataType, prep: Any) -> Selected:
    if isinstance(prep, str):
        # XXX: Backwards compatibility thing.
        #      str values are interpreted as Bind values and Bind values are
        #      assumed to be properties. So here we skip
        #      `env.call('select', prep)` and return `prep` as is.
        #      This should be eventually removed, once backwards compatibility
        #      for resolving strings as properties is removed.
        return Selected(prop=dtype.prop, prep=prep)
    elif isinstance(prep, Expr):
        # If `prepare` expression returns another expression, then this means,
        # it must be processed on values returned by query.
        prop = dtype.prop
        if not isinstance(env.backend, ExternalBackend) or (prop.external and prop.external.name):
            sel = env.call('select', dtype)
            return Selected(item=sel.item, prop=sel.prop, prep=prep)
        else:
            return Selected(item=None, prop=prop, prep=prep)
    else:
        result = env.call('select', prep)
        return Selected(prop=dtype.prop, prep=result)


@ufunc.resolver(QueryBuilder, Property, Func)
def select(
    env: QueryBuilder,
    prop: Property,
    func: Func
) -> Selected:
    return env.call('select', prop.dtype, func)


@ufunc.resolver(QueryBuilder, Flip)
def select(env: QueryBuilder, flip_: Flip):
    return env.call('select', flip_.prop, flip_)


@ufunc.resolver(QueryBuilder, Property, Flip)
def select(env: QueryBuilder, prop: Property, flip_: Flip) -> Selected:
    if flip_.required:
        return env.call('select', prop.dtype, flip_)
    # Skip `Property`, no need to resolve `external.prepare`, since `flip(...)` resolves it
    return env.call('select', prop.dtype)


@ufunc.resolver(QueryBuilder, ForeignProperty, Flip)
def select(
    env: QueryBuilder,
    fpr: ForeignProperty,
    flip_: Flip
) -> Selected:
    if flip_.required:
        return env.call('select', fpr, fpr.right.prop, flip_)
    # Skip `Property`, no need to resolve `external.prepare`, since `flip(...)` resolves it
    return env.call('select', fpr, fpr.right.prop.dtype)


@ufunc.resolver(QueryBuilder, DataType, Flip)
def select(env: QueryBuilder, dtype: DataType, flip_: Flip):
    return env.call('select', dtype, Expr('flip'))


@ufunc.resolver(QueryBuilder, Denorm, Flip)
def select(env: QueryBuilder, dtype: Denorm, flip_: Flip):
    fpr = env.call('_denorm_to_foreign_property', dtype)
    return env.call('select', fpr, flip_)


@ufunc.resolver(QueryBuilder, ForeignProperty, DataType, Flip)
def select(env: QueryBuilder, fpr: ForeignProperty, dtype: DataType, flip_: Flip):
    return env.call('select', fpr, dtype, Expr('flip'))


@ufunc.resolver(QueryBuilder, GetAttr)
def sort(env, field):
    dtype = env.resolve_property(field)
    return env.call('sort', dtype)


@ufunc.resolver(QueryBuilder, Expr, name='paginate')
def paginate(env, expr):
    if len(expr.args) != 1:
        raise InvalidArgumentInExpression(arguments=expr.args, expr='paginate')
    page = expr.args[0]
    if isinstance(page, Page):
        if page.enabled:
            if not page.filter_only:
                env.page.select = env.call('select', page)
                for by, page_by in page.by.items():
                    sorted_ = env.call('sort',
                                       Negative(page_by.prop.name) if by.startswith("-") else Bind(page_by.prop.name))
                    if sorted_ is not None:
                        if isinstance(sorted_, (list, set, tuple)):
                            env.page.sort += sorted_
                        else:
                            env.page.sort.append(sorted_)
            env.page.page_ = page
            env.page.size = page.size
            return env.resolve(get_pagination_compare_query(page))
    else:
        raise InvalidArgumentInExpression(arguments=expr.args, expr='paginate')


@ufunc.resolver(QueryBuilder, Expr, name='expand')
def expand(env, expr):
    result = []
    if expr.args:
        for arg in expr.args:
            resolved = env.resolve(arg)
            prop = env.resolve_property(resolved)
            result.append(prop)
        return result
    return None


@ufunc.resolver(QueryBuilder, str, name='op')
def op_(env, arg: str):
    if arg == '*':
        return Star()
    else:
        raise NotImplementedError


@ufunc.resolver(QueryBuilder, Expr, name='page')
def page_(env, expr):
    pass


@ufunc.resolver(QueryBuilder, DataType, list)
def validate_dtype_for_select(env, dtype: DataType, selected_props: List[Property]):
    raise NotImplemented(f"validate_dtype_for_select with {dtype.name} is not implemented")


@ufunc.resolver(QueryBuilder, Text, list)
def validate_dtype_for_select(env, dtype: Text, selected_props: List[Property]):
    for prop in selected_props:
        if dtype.prop.name == prop.name or prop.parent == dtype.prop:
            raise CannotSelectTextAndSpecifiedLang(dtype)


@ufunc.resolver(QueryBuilder, String, list)
def validate_dtype_for_select(env, dtype: String, selected_props: List[Property]):
    if dtype.prop.parent and isinstance(dtype.prop.parent.dtype, Text):
        parent = dtype.prop.parent
        for prop in selected_props:
            if parent.name == prop.name or prop == dtype.prop:
                raise CannotSelectTextAndSpecifiedLang(parent.dtype)


@ufunc.resolver(QueryBuilder, NestedProperty, object, names=COMPARE)
def compare(env: QueryBuilder, op: str, nested: NestedProperty, value: object):
    return env.call(op, nested.right, value)


@ufunc.resolver(QueryBuilder, Bind, object, names=COMPARE)
def compare(env, op, field, value):
    prop = env.resolve_property(field)
    return env.call(op, prop.dtype, value)


@ufunc.resolver(QueryBuilder, Expr)
def testlist(env: QueryBuilder, expr: Expr) -> tuple:
    args, kwargs = expr.resolve(env)
    result = []
    for arg in args:
        result.append(process_literal_value(arg))
    return tuple(result)


@ufunc.resolver(QueryBuilder)
def flip(env: QueryBuilder):
    return env.call('flip', env.this)


@ufunc.resolver(QueryBuilder, Property)
def flip(env: QueryBuilder, prop: Property):
    return env.call('flip', prop.dtype)


@ufunc.resolver(QueryBuilder, DataType)
def flip(env: QueryBuilder, dtype: DataType):
    return Flip(dtype.prop)


@ufunc.resolver(QueryBuilder, Expr)
def flip(env: QueryBuilder, expr: Expr):
    args, kwargs = expr.resolve(env)
    return env.call('flip', *args, **kwargs)


@ufunc.resolver(QueryBuilder, (GetAttr, Bind))
def flip(env: QueryBuilder, bind: GetAttr | Bind):
    prop = env.resolve_property(bind)
    prop = env.call('_resolve_inherited_flip', prop)
    flip_ = env.call('_resolve_flip', prop)
    flip_.prop = prop
    return flip_


@ufunc.resolver(QueryBuilder, Flip)
def flip(env: QueryBuilder, flip_: Flip):
    return Flip(flip_.prop, flip_.count + 1)


@ufunc.resolver(QueryBuilder, Property)
def _resolve_flip(env: QueryBuilder, prop: Property):
    resolved = prop
    if prop.external and prop.external.prepare:
        resolved = env(this=prop).resolve(prop.external.prepare)
    resolved = env.call('_resolve_flip', prop.dtype, resolved)
    return env.call('flip', resolved)


@ufunc.resolver(QueryBuilder, Property)
def _resolve_inherited_flip(env: QueryBuilder, prop: Property):
    return env.call('_resolve_inherited_flip', prop.dtype)


@ufunc.resolver(QueryBuilder, ForeignProperty)
def _resolve_inherited_flip(env: QueryBuilder, prop: ForeignProperty):
    return prop


@ufunc.resolver(QueryBuilder, DataType)
def _resolve_inherited_flip(env: QueryBuilder, dtype: DataType):
    return dtype.prop


@ufunc.resolver(QueryBuilder, Denorm)
def _resolve_inherited_flip(env: QueryBuilder, dtype: Denorm):
    return env.call('_denorm_to_foreign_property', dtype)


@ufunc.resolver(QueryBuilder, DataType, object)
def _resolve_flip(env: QueryBuilder, dtype: DataType, obj: object):
    return obj


@ufunc.resolver(QueryBuilder, Denorm, object)
def _resolve_flip(env: QueryBuilder, dtype: Denorm, obj: object):
    resolved = obj
    prop = dtype.rel_prop
    if prop.external and prop.external.prepare:
        resolved = env(this=prop).resolve(prop.external.prepare)

    return resolved


@ufunc.resolver(QueryBuilder, Denorm, Flip)
def _resolve_flip(env: QueryBuilder, dtype: Denorm, flip_: Flip):
    resolved = flip_
    prop = dtype.rel_prop
    if prop.external and prop.external.prepare:
        resolved = env(this=prop).resolve(prop.external.prepare)

        if isinstance(resolved, Flip):
            return Flip(dtype.prop, flip_.count + resolved.count)
    return resolved


@ufunc.resolver(QueryBuilder, ForeignProperty)
def _resolve_flip(env: QueryBuilder, prop: ForeignProperty):
    resolved = prop.right.prop
    if resolved.external and resolved.external.prepare:
        resolved = env(this=resolved).resolve(resolved.external.prepare)
    result: Flip = env.call('flip', resolved)
    result.prop = prop
    return result


@ufunc.resolver(QueryBuilder, Expr)
def swap(env: QueryBuilder, expr: Expr):
    args, kwargs = expr.resolve(env)
    return Expr('swap', *args, **kwargs)


@ufunc.resolver(QueryBuilder, Expr)
def split(env: QueryBuilder, expr: Expr):
    args, kwargs = expr.resolve(env)
    return Expr('split', *args, **kwargs)



from __future__ import annotations

import datetime
import uuid
from typing import Union, Any

import geoalchemy2.functions
import sqlalchemy as sa
from sqlalchemy.dialects.postgresql import UUID

from spinta import exceptions, commands
from spinta.auth import authorized
from spinta.backends.constants import TableType, BackendFeatures
from spinta.backends.postgresql.ufuncs.query.components import PgQueryBuilder, InheritForeignProperty, Lower, Recurse, \
    Negative, Positive
from spinta.components import Property, Action, Page
from spinta.core.ufuncs import Bind, Negative as Negative_
from spinta.core.ufuncs import Expr
from spinta.core.ufuncs import ufunc, GetAttr
from spinta.datasets.backends.sql.ufuncs.components import Selected
from spinta.exceptions import EmptyStringSearch, NoneValueComparison, NotImplementedFeature
from spinta.exceptions import FieldNotInResource
from spinta.types.datatype import Array
from spinta.types.datatype import DataType, ExternalRef, Inherit, BackRef, Time, ArrayBackRef, Denorm
from spinta.types.datatype import Date
from spinta.types.datatype import DateTime
from spinta.types.datatype import File
from spinta.types.datatype import Integer
from spinta.types.datatype import Number
from spinta.types.datatype import Object
from spinta.types.datatype import PrimaryKey
from spinta.types.datatype import Ref
from spinta.types.datatype import String
from spinta.types.datatype import UUID as UUID_dtype
from spinta.types.geometry.components import Geometry
from spinta.types.text.components import Text
from spinta.types.text.helpers import determine_language_property_for_text
from spinta.ufuncs.components import ForeignProperty
from spinta.ufuncs.querybuilder.components import ReservedProperty, \
    NestedProperty, ResultProperty, Flip
from spinta.ufuncs.querybuilder.helpers import get_column_with_extra, get_language_column, \
    expanded
from spinta.ufuncs.querybuilder.ufuncs import Star
from spinta.utils.data import take


def _gather_selected_properties(env: PgQueryBuilder):
    result = []
    if env.selected:
        for selected in env.selected.values():
            if selected and selected.prop:
                result.append(selected.prop)
    return result


def _get_property_for_select(env: PgQueryBuilder, name: str):
    prop = env.model.properties.get(name)
    if prop and authorized(env.context, prop, Action.SEARCH):
        return prop
    else:
        raise FieldNotInResource(env.model, property=name)


@ufunc.resolver(PgQueryBuilder, Expr)
def select(env, expr):
    keys = [str(k) for k in expr.args]
    args, kwargs = expr.resolve(env)
    args = list(zip(keys, args)) + list(kwargs.items())

    env.selected = {}
    if args:
        for key, arg in args:
            selected = env.call('select', arg)
            if selected is not None:
                env.selected[key] = selected
    else:
        env.call('select', Star())

    if not (len(args) == 1 and args[0][0] == '_page'):
        assert env.selected, args


@ufunc.resolver(PgQueryBuilder, Star)
def select(env, arg: Star) -> None:
    for prop in take(env.model.properties).values():
        # if authorized(env.context, prop, (Action.GETALL, Action.SEARCH)):
        # TODO: This line above should come from a getall(request),
        #       because getall can be used internally for example for
        #       writes.

        # Check if prop is expanded or not
        if not expanded(env, prop):
            continue
        env.selected[prop.place] = env.call('select', prop)


@ufunc.resolver(PgQueryBuilder, Bind)
def select(env, arg):
    if arg.name == '_type':
        return Selected(None, env.model.properties['_type'])
    if arg.name == '_page':
        return None
    prop = _get_property_for_select(env, arg.name)
    if not expanded(env, prop):
        return Selected(None, prop, prep=[])
    return env.call('select', prop)


@ufunc.resolver(PgQueryBuilder, ForeignProperty, DataType)
def select(
    env: PgQueryBuilder,
    fpr: ForeignProperty,
    dtype: DataType,
) -> Selected:
    table = env.get_joined_table_from_ref(fpr)
    column = env.backend.get_column(table, dtype.prop, select=True).label(
        fpr.left.prop.place + '.' + dtype.prop.name
    )
    return Selected(
        item=env.add_column(column),
        prop=dtype.prop,
    )


@ufunc.resolver(PgQueryBuilder, ForeignProperty, Ref)
def select(
    env: PgQueryBuilder,
    fpr: ForeignProperty,
    dtype: Ref,
) -> Selected:
    table = env.get_joined_table_from_ref(fpr)

    uri = dtype.model.uri_prop
    prep = {}
    if not dtype.inherited:
        name = '_id'
        if env.query_params.prioritize_uri and uri is not None:
            column = table.c[uri.place]
            name = '_uri'
            column = column.label(dtype.prop.place + '._uri')
        else:
            column = table.c[dtype.prop.place + '._id']
        column = env.add_column(column)
        prep[name] = Selected(column, dtype.prop)
    for prop in dtype.properties.values():
        sel = env.call('select', fpr.push(prop))
        prep[prop.name] = sel
    return Selected(prop=dtype.prop, prep=prep)


@ufunc.resolver(PgQueryBuilder, ForeignProperty, ExternalRef)
def select(
    env: PgQueryBuilder,
    fpr: ForeignProperty,
    dtype: Ref,
) -> Selected:
    # FIXME
    # Currently there are issues with resolving joins with Foreign properties
    # so we just skip it, this can cause issues with nested joins
    return env.call('select', dtype)


@ufunc.resolver(PgQueryBuilder, Property)
def select(env, prop):
    if prop.place not in env.resolved:
        if prop.external and prop.external.prepare:
            result = env(this=prop).resolve(prop.external.prepare)
            result = env.call("select", prop.dtype, result)
        else:
            result = env.call("select", prop.dtype)
        env.resolved[prop.place] = result
    return env.resolved[prop.place]


@ufunc.resolver(PgQueryBuilder, DataType)
def select(env: PgQueryBuilder, dtype: DataType):
    table = env.backend.get_table(env.model)

    if dtype.prop.list is None:
        column = env.backend.get_column(table, dtype.prop, select=True)
    else:
        # XXX: Probably if dtype.prop is in a nested list, we need to get the
        #      first list. Because if I remember correctly, only first list is
        #      stored on the main table.
        column = env.backend.get_column(table, dtype.prop.list, select=True)
    return Selected(env.add_column(column), dtype.prop)


@ufunc.resolver(PgQueryBuilder, String)
def select(env, dtype):
    env.call('validate_dtype_for_select', dtype, _gather_selected_properties(env))
    if dtype.prop.list is None:
        column = env.backend.get_column(env.table, dtype.prop)
    else:
        column = env.backend.get_column(env.table, dtype.prop.list)
    return Selected(env.add_column(column), dtype.prop)


@ufunc.resolver(PgQueryBuilder, Text)
def select(env, dtype):
    env.call('validate_dtype_for_select', dtype, _gather_selected_properties(env))
    if dtype.prop.list:
        column = env.backend.get_column(env.table, dtype.prop.list)
        return Selected(env.add_column(column), dtype.prop)

    if env.query_params.push:
        result = {
            key: env.call('select', prop)
            for key, prop in dtype.langs.items()
        }
        return Selected(prop=dtype.prop, prep=result)

    if env.query_params.lang:
        for lang in env.query_params.lang:
            if isinstance(lang, Star):
                result = {
                    key: env.call('select', prop)
                    for key, prop in dtype.langs.items()
                }
                return Selected(prop=dtype.prop, prep=result)
            break

        result = {
            key: env.call('select', prop)
            for key, prop in dtype.langs.items() if key in env.query_params.lang
        }
        return Selected(prop=dtype.prop, prep=result)

    default_langs = env.context.get('config').languages
    lang_prop = determine_language_property_for_text(dtype, env.query_params.lang_priority, default_langs)
    return Selected(prop=dtype.prop, prep={
        lang_prop.name: env.call('select', lang_prop),
    })


@ufunc.resolver(PgQueryBuilder, Object)
def select(env, dtype):
    prep = {}
    for prop in take(dtype.properties).values():
        sel = env.call('select', prop)
        if sel is not None:
            prep[prop.name] = sel
    return Selected(prop=dtype.prop, prep=prep)


@ufunc.resolver(PgQueryBuilder, File)
def select(env, dtype):
    table = env.backend.get_table(env.model)
    name_id = dtype.prop.place + '._id'
    name_content_type = dtype.prop.place + '._content_type'
    name_size = dtype.prop.place + '._size'
    prep = {
        '_id': Selected(env.add_column(table.c[name_id])),
        '_content_type': Selected(env.add_column(table.c[name_content_type])),
        '_size': Selected(env.add_column(table.c[name_size]))
    }
    same_backend = dtype.backend.name == dtype.prop.model.backend.name
    if same_backend or BackendFeatures.FILE_BLOCKS in dtype.backend.features:
        name_bsize = dtype.prop.place + '._bsize'
        name_blocks = dtype.prop.place + '._blocks'
        prep['_bsize'] = Selected(env.add_column(table.c[name_bsize]))
        prep['_blocks'] = Selected(env.add_column(table.c[name_blocks]))
    return Selected(prop=dtype.prop, prep=prep)


@ufunc.resolver(PgQueryBuilder, File, str)
def select(env, dtype, leaf):
    table = env.backend.get_table(env.model)
    if leaf == '_content':
        # Currently _content is handled in prepare_dtype_for_response
        # we need to make sure that whole File is loaded when _content is selected
        env.selected[dtype.prop.place] = env.call('select', dtype.prop)
    else:
        column = table.c[dtype.prop.place + '.' + leaf]
        return Selected(env.add_column(column), dtype.prop)


@ufunc.resolver(PgQueryBuilder, Ref)
def select(env, dtype: Ref):
    uri = dtype.model.uri_prop
    prep = {}
    if not dtype.inherited:
        name = '_id'
        if env.query_params.prioritize_uri and uri is not None:
            fpr = ForeignProperty(None, dtype, dtype.model.properties['_id'].dtype)
            table = env.get_joined_table(fpr)
            column = table.c[uri.place]
            name = '_uri'
            column = column.label(dtype.prop.place + '._uri')
        else:
            table = env.backend.get_table(env.model)
            column = table.c[dtype.prop.place + '._id']
        column = env.add_column(column)
        prep[name] = Selected(column, dtype.prop)
    for prop in dtype.properties.values():
        sel = env.call('select', prop)
        prep[prop.name] = sel
    return Selected(prop=dtype.prop, prep=prep)


@ufunc.resolver(PgQueryBuilder, ExternalRef)
def select(env, dtype: ExternalRef):
    prep = {}
    if not dtype.inherited:
        table = env.backend.get_table(env.model)
        if dtype.model.given.pkeys or dtype.explicit:
            props = dtype.refprops
        else:
            props = [dtype.model.properties['_id']]
        for prop in props:
            name = f"{dtype.prop.place}.{prop.place}"
            column = table.c[name]
            prep[prop.name] = Selected(env.add_column(column), prop)
    for prop in dtype.properties.values():
        sel = env.call('select', prop)
        prep[prop.name] = sel
    return Selected(prop=dtype.prop, prep=prep)


@ufunc.resolver(PgQueryBuilder, Ref, str)
def select(env, dtype: Ref, prop: str):
    table = env.backend.get_table(env.model)
    return Selected(
        env.add_column(
            table.c[dtype.prop.place + '.' + prop]
        )
    )


@ufunc.resolver(PgQueryBuilder, BackRef)
def select(env, dtype):
    return _select_backref(env, dtype)


@ufunc.resolver(PgQueryBuilder, ArrayBackRef)
def select(env: PgQueryBuilder, dtype: ArrayBackRef):
    selected = _select_backref(env, dtype.items.dtype, is_array=True)
    return Selected(prop=dtype.prop, prep=selected.prep)


def _select_backref(env, dtype, is_array=False):
    fpr = ForeignProperty(
        None,
        left=dtype,
        right=dtype.refprop.dtype,
    )
    refprop = dtype.refprop
    required_columns = []
    return_columns = {}

    if commands.identifiable(refprop):
        id_ = fpr.right.prop.model.properties['_id']
        column_name = id_.name
        label = f'{dtype.prop.name}.{column_name}'
        required_columns.append((column_name, label))
        return_columns[label] = id_
    else:
        for prop in dtype.refprop.dtype.refprops:
            column_name = prop.name
            label = f'{dtype.prop.name}.{column_name}'
            required_columns.append((column_name, label))
            return_columns[label] = prop

    selector = env.generate_backref_select(dtype.prop, dtype.refprop, required_columns, is_array)
    table = env.get_backref_joined_table(fpr, selector)

    prep = {}

    if is_array:
        key = dtype.prop.name
        prop = dtype.prop
        selected_column = env.add_column(table.c[key])
        return Selected(prop=prop, prep=Selected(selected_column, prop))

    for key, prop in return_columns.items():
        prep[prop.place] = Selected(env.add_column(table.c[key]), prop)
    return Selected(prop=fpr.left.prop, prep=prep)


@ufunc.resolver(PgQueryBuilder, ForeignProperty, Inherit)
def select(
    env: PgQueryBuilder,
    fpr: ForeignProperty,
    dtype: Inherit,
) -> Selected:
    table = env.get_joined_table(fpr)
    fixed_name = dtype.prop.place
    if fixed_name.startswith(f'{fpr.left.prop.place}.'):
        fixed_name = fixed_name.replace(f'{fpr.left.prop.place}.', '', 1)
    column = table.c[fixed_name]
    column = column.label(fpr.place)
    return Selected(env.add_column(column), dtype.prop)


@ufunc.resolver(PgQueryBuilder, ForeignProperty, Denorm)
def select(
    env: PgQueryBuilder,
    fpr: ForeignProperty,
    dtype: Denorm,
) -> Selected:
    return env.call('select', fpr.swap(dtype.rel_prop))


@ufunc.resolver(PgQueryBuilder, Inherit)
def select(env, dtype):
    table = env.get_joined_base_table(dtype.prop.model, dtype.prop.name)
    column = table.c[dtype.prop.name]
    column = column.label(dtype.prop.name)
    return Selected(env.add_column(column), dtype.prop)


@ufunc.resolver(PgQueryBuilder, InheritForeignProperty)
def select(env, dtype):
    table = env.get_joined_base_table(dtype.model, dtype.prop_name)
    column = table.c[dtype.prop_name]
    column = column.label(f"{dtype.base_prop.name}.{dtype.prop_name}")
    return Selected(env.add_column(column), dtype.base_prop)


@ufunc.resolver(PgQueryBuilder, Denorm)
def _denorm_to_foreign_property(
    env: PgQueryBuilder,
    dtype: Denorm
):
    ref = dtype.prop.parent
    if isinstance(ref, Property) and isinstance(ref.dtype, (Ref, BackRef)):
        fpr = None
        if ref.dtype.inherited:
            root_ref_parent = ref.parent

            while root_ref_parent and isinstance(root_ref_parent, Property) and isinstance(root_ref_parent.dtype, (Ref, BackRef)):
                fpr = ForeignProperty(fpr, root_ref_parent.dtype, root_ref_parent.dtype.model.properties['_id'].dtype)

                if not root_ref_parent.dtype.inherited:
                    break
                root_ref_parent = root_ref_parent.parent
        return ForeignProperty(fpr, ref.dtype, dtype.rel_prop.dtype)
    return dtype.rel_prop


@ufunc.resolver(PgQueryBuilder, Denorm)
def select(env, dtype):
    fpr = env.call('_denorm_to_foreign_property', dtype)
    result = env.call('select', fpr)
    return result


@ufunc.resolver(PgQueryBuilder, Page)
def select(env, page):
    return_selected = []
    for item in page.by.values():
        selected = env.call('select', item.prop)
        if selected.item is not None:
            return_selected.append(env.columns[selected.item])
    return return_selected


@ufunc.resolver(PgQueryBuilder, sa.sql.expression.ColumnElement)
def select(env, column):
    return Selected(env.add_column(column))


@ufunc.resolver(PgQueryBuilder, ForeignProperty, PrimaryKey)
def select(
    env: PgQueryBuilder,
    fpr: ForeignProperty,
    dtype: PrimaryKey,
) -> Selected:
    super_ = ufunc.resolver[env, fpr, dtype]
    return super_(env, fpr, dtype)


@ufunc.resolver(PgQueryBuilder, Geometry, Flip)
def select(env: PgQueryBuilder, dtype: Geometry, func_: Flip):
    table = env.backend.get_table(env.model)

    if dtype.prop.list is None:
        column = env.backend.get_column(table, dtype.prop, select=True)
    else:
        column = env.backend.get_column(table, dtype.prop.list, select=True)

    column = geoalchemy2.functions.ST_FlipCoordinates(column)
    return Selected(env.add_column(column), prop=dtype.prop)


@ufunc.resolver(PgQueryBuilder, ForeignProperty, Geometry, Flip)
def select(env: PgQueryBuilder, fpr: ForeignProperty, dtype: Geometry, func_: Flip):
    table = env.get_joined_table(fpr)

    if dtype.prop.list is None:
        column = env.backend.get_column(table, dtype.prop, select=True)
    else:
        column = env.backend.get_column(table, dtype.prop.list, select=True)

    column = geoalchemy2.functions.ST_FlipCoordinates(column)
    return Selected(env.add_column(column), prop=dtype.prop)


@ufunc.resolver(PgQueryBuilder, int)
def limit(env, n):
    env.limit = n


@ufunc.resolver(PgQueryBuilder, int)
def offset(env, n):
    env.offset = n


@ufunc.resolver(PgQueryBuilder, Expr, name='and')
def and_(env, expr):
    args, kwargs = expr.resolve(env)
    args = [a for a in args if a is not None]
    if len(args) > 1:
        return sa.and_(*args)
    elif args:
        return args[0]


@ufunc.resolver(PgQueryBuilder, Expr, name='or')
def or_(env, expr):
    args, kwargs = expr.resolve(env)
    args = [a for a in args if a is not None]
    return env.call('or', args)


@ufunc.resolver(PgQueryBuilder, list, name='or')
def or_(env, args):
    if len(args) > 1:
        return sa.or_(*args)
    elif args:
        return args[0]


@ufunc.resolver(PgQueryBuilder)
def count(env):
    env.aggregate = True
    return sa.func.count().label('count()')


COMPARE = [
    'eq',
    'ne',
    'lt',
    'le',
    'gt',
    'ge',
    'startswith',
    'contains',
]

COMPARE_EQUATIONS = [
    'eq',
    'ne',
    'lt',
    'le',
    'gt',
    'ge',
]

COMPARE_STRING = [
    'eq',
    'startswith',
    'contains',
]


@ufunc.resolver(PgQueryBuilder, GetAttr, object, names=COMPARE)
def compare(env: PgQueryBuilder, op: str, attr: GetAttr, value: Any):
    resolved = env.resolve_property(attr)
    return env.call(op, resolved, value)


@ufunc.resolver(PgQueryBuilder, ReservedProperty, object, names=COMPARE)
def compare(env, op: str, reserved: ReservedProperty, value: Any):
    table = env.backend.get_table(reserved.dtype.prop.model)
    column = table.c[reserved.dtype.prop.place + '.' + reserved.param]

    cond = _sa_compare(op, column, value)
    return _prepare_condition(env, reserved.dtype.prop, cond)


@ufunc.resolver(PgQueryBuilder, PrimaryKey, object, names=COMPARE_EQUATIONS)
def compare(env, op, dtype, value):
    column = env.backend.get_column(env.table, dtype.prop)
    cond = _sa_compare(op, column, value)
    return _prepare_condition(env, dtype.prop, cond)


@ufunc.resolver(PgQueryBuilder, ForeignProperty, object, names=COMPARE)
def compare(env, op: str, fpr: ForeignProperty, value: Any):
    return env.call(op, fpr, fpr.right, value)


@ufunc.resolver(PgQueryBuilder, PrimaryKey, object, names=COMPARE)
def compare(env, op, dtype, value):
    str_value = str(value)
    try:
        uuid.UUID(str_value)
    except ValueError:
        raise exceptions.InvalidValue(dtype, op=op, arg=type(value).__name__)

    column = env.backend.get_column(env.table, dtype.prop)
    cond = _sa_compare(op, column, str_value)
    return _prepare_condition(env, dtype.prop, cond)


@ufunc.resolver(PgQueryBuilder, PrimaryKey, type(None), names=COMPARE)
def compare(env, op, dtype, value):
    column = env.backend.get_column(env.table, dtype.prop)
    cond = _sa_compare(op, column, value)
    return _prepare_condition(env, dtype.prop, cond)


@ufunc.resolver(PgQueryBuilder, UUID_dtype, str, names=COMPARE)
def compare(env, op, dtype, value):
    column = env.backend.get_column(env.table, dtype.prop)
    cond = _sa_compare(op, column, value)
    return _prepare_condition(env, dtype.prop, cond)


@ufunc.resolver(PgQueryBuilder, String, str, names=COMPARE)
def compare(env, op, dtype, value):
    if op in ('startswith', 'contains'):
        _ensure_non_empty(op, value)
    column = env.backend.get_column(env.table, dtype.prop)
    cond = _sa_compare(op, column, value)
    return _prepare_condition(env, dtype.prop, cond)


@ufunc.resolver(PgQueryBuilder, (Integer, Number), (int, float), names=COMPARE_EQUATIONS)
def compare(env, op, dtype, value):
    column = env.backend.get_column(env.table, dtype.prop)
    cond = _sa_compare(op, column, value)
    return _prepare_condition(env, dtype.prop, cond)


@ufunc.resolver(PgQueryBuilder, DateTime, str, names=COMPARE_EQUATIONS)
def compare(env, op, dtype, value):
    column = env.backend.get_column(env.table, dtype.prop)
    value = datetime.datetime.fromisoformat(value)
    cond = _sa_compare(op, column, value)
    return _prepare_condition(env, dtype.prop, cond)


@ufunc.resolver(PgQueryBuilder, DateTime, datetime.datetime, names=COMPARE_EQUATIONS)
def compare(env, op, dtype, value):
    column = env.backend.get_column(env.table, dtype.prop)
    cond = _sa_compare(op, column, value)
    return _prepare_condition(env, dtype.prop, cond)


@ufunc.resolver(PgQueryBuilder, Date, str, names=COMPARE_EQUATIONS)
def compare(env, op, dtype, value):
    column = env.backend.get_column(env.table, dtype.prop)
    value = datetime.date.fromisoformat(value)
    cond = _sa_compare(op, column, value)
    return _prepare_condition(env, dtype.prop, cond)


@ufunc.resolver(PgQueryBuilder, Date, datetime.date, names=COMPARE_EQUATIONS)
def compare(env, op, dtype, value):
    column = env.backend.get_column(env.table, dtype.prop)
    cond = _sa_compare(op, column, value)
    return _prepare_condition(env, dtype.prop, cond)


@ufunc.resolver(PgQueryBuilder, Time, str, names=COMPARE_EQUATIONS)
def compare(env, op, dtype, value):
    column = env.backend.get_column(env.table, dtype.prop)
    value = datetime.time.fromisoformat(value)
    cond = _sa_compare(op, column, value)
    return _prepare_condition(env, dtype.prop, cond)


@ufunc.resolver(PgQueryBuilder, Time, datetime.time, names=COMPARE_EQUATIONS)
def compare(env, op, dtype, value):
    column = env.backend.get_column(env.table, dtype.prop)
    cond = _sa_compare(op, column, value)
    return _prepare_condition(env, dtype.prop, cond)


@ufunc.resolver(PgQueryBuilder, DataType, object, names=COMPARE)
def compare(env, op, dtype, value):
    raise exceptions.InvalidValue(dtype, op=op, arg=type(value).__name__)


@ufunc.resolver(PgQueryBuilder, ForeignProperty, DataType, object, names=COMPARE)
def compare(
    env: PgQueryBuilder,
    op: str,
    fpr: ForeignProperty,
    dtype: DataType,
    value: Any,
):
    raise exceptions.InvalidValue(dtype, op=op, arg=type(value).__name__)


@ufunc.resolver(PgQueryBuilder, UUID_dtype, str)
def eq(env, dtype, value):
    column = env.backend.get_column(env.table, dtype.prop)
    cond = _sa_compare('eq', column, value)
    return _prepare_condition(env, dtype.prop, cond)


@ufunc.resolver(PgQueryBuilder, DataType, type(None))
def eq(env, dtype, value):
    column = env.backend.get_column(env.table, dtype.prop)
    cond = _sa_compare('eq', column, value)
    return _prepare_condition(env, dtype.prop, cond)


@ufunc.resolver(PgQueryBuilder, Text, (str, Bind, type(None)))
def eq(env, dtype, value):
    column = get_language_column(env, dtype.prop)
    cond = _sa_compare('eq', column, value)
    return _prepare_condition(env, dtype.prop, cond)


@ufunc.resolver(PgQueryBuilder, ForeignProperty, DataType, type(None))
def eq(
    env: PgQueryBuilder,
    fpr: ForeignProperty,
    dtype: DataType,
    value: type(None),
):
    table = env.get_joined_table(fpr)
    column = env.backend.get_column(table, fpr.right.prop)
    cond = _sa_compare('eq', column, value)
    return _prepare_condition(env, dtype.prop, cond)


@ufunc.resolver(PgQueryBuilder, ForeignProperty, PrimaryKey, str)
def eq(
    env: PgQueryBuilder,
    fpr: ForeignProperty,
    dtype: DataType,
    value: type(None),
):
    table = env.backend.get_table(fpr.left.model)
    column = env.backend.get_column(table, fpr.left.prop)
    cond = _sa_compare('eq', column, value)
    return _prepare_condition(env, dtype.prop, cond)


def _ensure_non_empty(op, s):
    if s == '':
        raise EmptyStringSearch(op=op)


@ufunc.resolver(PgQueryBuilder, UUID_dtype, str, names=COMPARE_STRING)
def compare(env: PgQueryBuilder, op: str, dtype: UUID, value: str):
    if op in ('startswith', 'contains'):
        _ensure_non_empty(op, value)
    column = env.backend.get_column(env.table, dtype.prop).cast(sa.String)
    return _sa_compare(op, column, value)


@ufunc.resolver(PgQueryBuilder, ForeignProperty, String, str, names=COMPARE_STRING)
def compare(
    env: PgQueryBuilder,
    op: str,
    fpr: ForeignProperty,
    dtype: String,
    value: str,
):
    if op in ('startswith', 'contains'):
        _ensure_non_empty(op, value)
    table = env.get_joined_table(fpr)
    column = table.c[fpr.right.prop.place]
    cond = _sa_compare(op, column, value)
    return _prepare_condition(env, dtype.prop, cond)


@ufunc.resolver(PgQueryBuilder, PrimaryKey, str, names=COMPARE_STRING)
def compare(env, op, dtype, value):
    if op in ('startswith', 'contains'):
        _ensure_non_empty(op, value)
    column = env.backend.get_column(env.table, dtype.prop)
    return _sa_compare(op, column, value)


@ufunc.resolver(PgQueryBuilder, ForeignProperty, (Integer, Number), (int, float), names=COMPARE_EQUATIONS)
def compare(
    env: PgQueryBuilder,
    op: str,
    fpr: ForeignProperty,
    dtype: Union[Integer, Number],
    value: Union[int, float],
):
    table = env.get_joined_table(fpr)
    column = table.c[fpr.right.prop.place]
    cond = _sa_compare(op, column, value)
    return _prepare_condition(env, dtype.prop, cond)


@ufunc.resolver(PgQueryBuilder, ForeignProperty, DateTime, str, names=COMPARE_EQUATIONS)
def compare(
    env: PgQueryBuilder,
    op: str,
    fpr: ForeignProperty,
    dtype: DateTime,
    value: str,
):
    table = env.get_joined_table(fpr)
    column = table.c[fpr.right.prop.place]
    value = datetime.datetime.fromisoformat(value)
    cond = _sa_compare(op, column, value)
    return _prepare_condition(env, dtype.prop, cond)


@ufunc.resolver(PgQueryBuilder, ForeignProperty, Date, str, names=COMPARE_EQUATIONS)
def compare(
    env: PgQueryBuilder,
    op: str,
    fpr: ForeignProperty,
    dtype: Date,
    value: str,
):
    table = env.get_joined_table(fpr)
    column = table.c[fpr.right.prop.place]
    value = datetime.date.fromisoformat(value)
    cond = _sa_compare(op, column, value)
    return _prepare_condition(env, dtype.prop, cond)


@ufunc.resolver(PgQueryBuilder, String)
def lower(env, dtype):
    return Lower(dtype)


@ufunc.resolver(PgQueryBuilder, Recurse)
def lower(env, recurse):
    return Recurse([env.call('lower', arg) for arg in recurse.args])


@ufunc.resolver(PgQueryBuilder, Lower, str, names=COMPARE_STRING)
def compare(env, op, fn, value):
    if op in ('startswith', 'contains'):
        _ensure_non_empty(op, value)
    column = get_column_with_extra(env, fn.dtype.prop)
    column = sa.func.lower(column)
    cond = _sa_compare(op, column, value)
    return _prepare_condition(env, fn.dtype.prop, cond)


def _sa_compare(op, column, value):
    if value is None and op not in ['eq', 'ne']:
        raise NoneValueComparison(op=op)

    # Convert JSONB value from -> to ->> with astext
    if isinstance(column.type, sa.JSON):
        if not isinstance(column, sa.Column):
            column = column.element
        column = column.astext

    if op == 'eq':
        return column == value

    if op == 'lt':
        return column < value

    if op == 'le':
        return column <= value

    if op == 'gt':
        return column > value

    if op == 'ge':
        return column >= value

    if op == 'contains':
        if isinstance(column.type, UUID):
            column = column.cast(sa.String)
        return column.contains(value)

    if op == 'startswith':
        if isinstance(column.type, UUID):
            column = column.cast(sa.String)
        return column.startswith(value)
    raise NotImplementedError


def _prepare_condition(env: PgQueryBuilder, prop: Property, cond):
    if prop.list is None:
        return cond

    main_table = env.table
    list_table = env.backend.get_table(prop.list, TableType.LIST)
    subqry = (
        sa.select(
            [list_table.c._rid],
            distinct=list_table.c._rid,
        ).
        where(cond).
        alias()
    )
    env.from_ = env.from_.outerjoin(
        subqry,
        main_table.c._id == subqry.c._rid,
    )
    return subqry.c._rid.isnot(None)


@ufunc.resolver(PgQueryBuilder, DataType, type(None))
def ne(env, dtype, value):
    column = env.backend.get_column(env.table, dtype.prop)
    return _ne_compare(env, dtype.prop, column, value)


@ufunc.resolver(PgQueryBuilder, UUID_dtype, str)
def ne(env, dtype, value):
    column = env.backend.get_column(env.table, dtype.prop)
    return _ne_compare(env, dtype.prop, column, value)


@ufunc.resolver(PgQueryBuilder, ForeignProperty, DataType, type(None))
def ne(
    env: PgQueryBuilder,
    fpr: ForeignProperty,
    dtype: DataType,
    value: type(None),
):
    table = env.get_joined_table(fpr)
    column = table.c[fpr.right.prop.place]
    return _ne_compare(env, dtype.prop, column, value)


@ufunc.resolver(PgQueryBuilder, String, str)
def ne(env, dtype, value):
    column = env.backend.get_column(env.table, dtype.prop)
    return _ne_compare(env, dtype.prop, column, value)


@ufunc.resolver(PgQueryBuilder, Text, str)
def ne(env, dtype, value):
    column = get_language_column(env, dtype)
    return _ne_compare(env, dtype.prop, column, value)


@ufunc.resolver(PgQueryBuilder, ForeignProperty, String, str)
def ne(
    env: PgQueryBuilder,
    fpr: ForeignProperty,
    dtype: DataType,
    value: str,
):
    table = env.get_joined_table(fpr)
    column = table.c[fpr.right.prop.place]
    return _ne_compare(env, dtype.prop, column, value)


@ufunc.resolver(PgQueryBuilder, PrimaryKey, str)
def ne(env, dtype, value):
    column = env.backend.get_column(env.table, dtype.prop)
    return _ne_compare(env, dtype.prop, column, value)


@ufunc.resolver(PgQueryBuilder, (Integer, Number), (int, float))
def ne(env, dtype, value):
    column = env.backend.get_column(env.table, dtype.prop)
    return _ne_compare(env, dtype.prop, column, value)


@ufunc.resolver(PgQueryBuilder, ForeignProperty, (Integer, Number), (int, float))
def ne(
    env: PgQueryBuilder,
    fpr: ForeignProperty,
    dtype: Union[Integer, Number],
    value: Union[int, float],
):
    table = env.get_joined_table(fpr)
    column = table.c[fpr.right.prop.place]
    return _ne_compare(env, dtype.prop, column, value)


@ufunc.resolver(PgQueryBuilder, DateTime, str)
def ne(env, dtype, value):
    column = env.backend.get_column(env.table, dtype.prop)
    value = datetime.datetime.fromisoformat(value)
    return _ne_compare(env, dtype.prop, column, value)


@ufunc.resolver(PgQueryBuilder, ForeignProperty, DateTime, str)
def ne(
    env: PgQueryBuilder,
    fpr: ForeignProperty,
    dtype: DateTime,
    value: str,
):
    table = env.get_joined_table(fpr)
    column = table.c[fpr.right.prop.place]
    value = datetime.datetime.fromisoformat(value)
    return _ne_compare(env, dtype.prop, column, value)


@ufunc.resolver(PgQueryBuilder, Date, str)
def ne(env, dtype, value):
    column = env.backend.get_column(env.table, dtype.prop)
    value = datetime.date.fromisoformat(value)
    return _ne_compare(env, dtype.prop, column, value)


@ufunc.resolver(PgQueryBuilder, ForeignProperty, Date, str)
def ne(
    env: PgQueryBuilder,
    fpr: ForeignProperty,
    dtype: Date,
    value: str,
):
    table = env.get_joined_table(fpr)
    column = table.c[fpr.right.prop.place]
    value = datetime.date.fromisoformat(value)
    return _ne_compare(env, dtype.prop, column, value)


@ufunc.resolver(PgQueryBuilder, Array, (object, type(None)), names=COMPARE)
def compare(env, op, dtype, value):
    return env.call(op, dtype.items.dtype, value)


@ufunc.resolver(PgQueryBuilder, Lower, str)
def ne(env, fn, value):
    column = get_column_with_extra(env, fn.dtype.prop)
    column = sa.func.lower(column)
    return _ne_compare(env, fn.dtype.prop, column, value)


def _ne_compare(env: PgQueryBuilder, prop: Property, column, value):
    """Not equal operator is quite complicated thing and need explaining.

    If property is not defined within a list, just do `!=` comparison and be
    done with it.

    If property is in a list:

    - First check if there is at least one list item where field is not None
        (existance check).

    - Then check if there is no list items where field equals to given
        value.
    """

    if prop.list is None:
        return column != value

    main_table = env.backend.get_table(prop.model)
    list_table = env.backend.get_table(prop.list, TableType.LIST)

    # Check if at liest one value for field is defined
    subqry1 = (
        sa.select(
            [list_table.c._rid],
            distinct=list_table.c._rid,
        ).
        where(column != None).  # noqa
        alias()
    )
    env.from_ = env.from_.outerjoin(
        subqry1,
        main_table.c._id == subqry1.c._rid,
    )

    # Check if given value exists
    subqry2 = (
        sa.select(
            [list_table.c._rid],
            distinct=list_table.c._rid,
        ).
        where(column == value).
        alias()
    )
    env.from_ = env.from_.outerjoin(
        subqry2,
        main_table.c._id == subqry2.c._rid,
    )

    # If field exists and given value does not, then field is not equal to
    # value.
    return sa.and_(
        subqry1.c._rid != None,  # noqa
        subqry2.c._rid == None,
    )


FUNCS = [
    'lower',
    'upper',
]


@ufunc.resolver(PgQueryBuilder, Bind, names=FUNCS)
def func(env, name, field):
    prop = env.resolve_property(field)
    return env.call(name, prop.dtype)


@ufunc.resolver(PgQueryBuilder, GetAttr, names=FUNCS)
def func(env, name, field):
    resolved = env.resolve_property(field)
    return env.call(name, resolved)


@ufunc.resolver(PgQueryBuilder, NestedProperty, names=FUNCS)
def func(env, name, nested):
    return env.call(name, nested.right)


@ufunc.resolver(PgQueryBuilder, Bind)
def recurse(env, field):
    if field.name in env.model.leafprops:
        return Recurse([prop.dtype for prop in env.model.leafprops[field.name]])
    else:
        raise exceptions.FieldNotInResource(env.model, property=field.name)


@ufunc.resolver(PgQueryBuilder, Recurse, object, names=COMPARE)
def recurse(env, op, recurse, value):
    return env.call('or', [
        env.call(op, arg, value)
        for arg in recurse.args
    ])


@ufunc.resolver(PgQueryBuilder, Expr, name='any')
def any_(env, expr):
    args, kwargs = expr.resolve(env)
    op, field, *args = args
    if isinstance(op, Bind):
        op = op.name
    return env.call('or', [
        env.call(op, field, arg)
        for arg in args
    ])


@ufunc.resolver(PgQueryBuilder, Expr)
def sort(env, expr):
    args, kwargs = expr.resolve(env)
    env.sort = [
        env.call('sort', arg) for arg in args
    ]


@ufunc.resolver(PgQueryBuilder, Bind)
def sort(env, field):
    prop = env.resolve_property(field)
    return env.call('asc', prop.dtype)


@ufunc.resolver(PgQueryBuilder, DataType)
def sort(env, dtype):
    return env.call('asc', dtype)


@ufunc.resolver(PgQueryBuilder, Negative_)
def sort(env, field):
    prop = env.resolve_property(field)
    return env.call('desc', prop.dtype)


@ufunc.resolver(PgQueryBuilder, ForeignProperty)
def sort(env: PgQueryBuilder, fpr: ForeignProperty):
    return env.call('asc', fpr, fpr.right)


@ufunc.resolver(PgQueryBuilder, Positive)
def sort(env, sign):
    return env.call('asc', sign.arg)


@ufunc.resolver(PgQueryBuilder, Negative)
def sort(env, sign):
    return env.call('desc', sign.arg)


@ufunc.resolver(PgQueryBuilder, DataType)
def asc(env, dtype):
    column = _get_sort_column(env, dtype.prop)
    return column.asc()


@ufunc.resolver(PgQueryBuilder, NestedProperty)
def asc(env: PgQueryBuilder, nested: NestedProperty):
    return env.call('asc', nested.right)


@ufunc.resolver(PgQueryBuilder, ForeignProperty)
def asc(env: PgQueryBuilder, fpr: ForeignProperty):
    return env.call('asc', fpr, fpr.right)


@ufunc.resolver(PgQueryBuilder, ForeignProperty, DataType)
def asc(env: PgQueryBuilder, fpr: ForeignProperty, dtype: DataType):
    table = env.get_joined_table(fpr)
    column = table.c[fpr.right.prop.place]
    return column.asc()


@ufunc.resolver(PgQueryBuilder, DataType)
def desc(env, dtype):
    column = _get_sort_column(env, dtype.prop)
    return column.desc()


@ufunc.resolver(PgQueryBuilder, NestedProperty)
def desc(env: PgQueryBuilder, nested: NestedProperty):
    return env.call('desc', nested.right)


@ufunc.resolver(PgQueryBuilder, ForeignProperty)
def desc(env: PgQueryBuilder, fpr: ForeignProperty):
    return env.call('desc', fpr, fpr.right)


@ufunc.resolver(PgQueryBuilder, ForeignProperty, DataType)
def desc(env: PgQueryBuilder, fpr: ForeignProperty, dtype: DataType):
    table = env.get_joined_table(fpr)
    column = table.c[fpr.right.prop.place]
    return column.desc()


@ufunc.resolver(PgQueryBuilder, Array, names=['asc', 'desc'])
def sort(env, name, dtype: Array):
    return env.call(name, dtype.items.dtype)


def _get_sort_column(env: PgQueryBuilder, prop: Property):
    column = get_column_with_extra(env, prop)

    if prop.list is None:
        return column

    main_table = env.table
    list_table = env.backend.get_table(prop.list, TableType.LIST)
    subqry = (
        sa.select(
            [list_table.c._rid, column.label('value')],
            distinct=list_table.c._rid,
        ).alias()
    )
    env.from_ = env.from_.outerjoin(
        subqry,
        main_table.c._id == subqry.c._rid,
    )
    return subqry.c.value


@ufunc.resolver(PgQueryBuilder, GetAttr)
def negative(env, field) -> Negative:
    resolved = env.resolve_property(field)
    return Negative(resolved)


@ufunc.resolver(PgQueryBuilder, GetAttr)
def positive(env, field) -> Positive:
    resolved = env.resolve_property(field)
    return Positive(resolved)


@ufunc.resolver(PgQueryBuilder, Bind)
def negative(env, field) -> Negative:
    prop = env.resolve_property(field)
    return Negative(prop.dtype)


@ufunc.resolver(PgQueryBuilder, Bind)
def positive(env, field) -> Positive:
    prop = env.resolve_property(field)
    return Positive(prop.dtype)


@ufunc.resolver(PgQueryBuilder, DataType)
def negative(env, dtype) -> Negative:
    return Negative(dtype)


@ufunc.resolver(PgQueryBuilder, ForeignProperty)
def negative(env: PgQueryBuilder, fpr: ForeignProperty) -> Negative:
    return Negative(fpr)


@ufunc.resolver(PgQueryBuilder, DataType)
def positive(env, dtype) -> Positive:
    return Positive(dtype)


@ufunc.resolver(PgQueryBuilder, ForeignProperty)
def positive(env: PgQueryBuilder, fpr: ForeignProperty) -> Positive:
    return Positive(fpr)


@ufunc.resolver(PgQueryBuilder, Expr)
def checksum(env: PgQueryBuilder, expr: Expr):
    args = []
    for prop in take(env.model.properties).values():
        if env.selected is not None and prop.place in env.selected:
            selected = env.selected[prop.place]
        else:
            selected = env.call('select', prop)
        if isinstance(selected, Selected):
            args.append(selected)
    return ResultProperty(
        Expr('checksum', *args)
    )


@ufunc.resolver(PgQueryBuilder, Expr)
def file(env: PgQueryBuilder, expr: Expr) -> Expr:
    raise NotImplementedFeature(env.backend, feature="Ability to use file() function with `PostgreSql` backend")


@ufunc.resolver(PgQueryBuilder, Bind, Bind)
def point(env: PgQueryBuilder, x: Bind, y: Bind) -> Expr:
    return Expr(
        'point',
        env.call('select', x),
        env.call('select', y),
    )



from __future__ import annotations

from spinta.datasets.backends.sql.ufuncs.query.components import SqlQueryBuilder
from spinta.datasets.backends.sql.backends.mysql.components import MySQL


class MySQLQueryBuilder(SqlQueryBuilder):
    backend: MySQL



import logging
from typing import Iterator

from spinta import commands
from spinta.backends.helpers import validate_and_return_begin
from spinta.components import Context, Property
from spinta.components import Model
from spinta.core.ufuncs import Expr
from spinta.datasets.backends.helpers import handle_ref_key_assignment, generate_pk_for_row, handle_external_array_type
from spinta.datasets.backends.sql.components import Sql
from spinta.datasets.helpers import get_enum_filters
from spinta.datasets.helpers import get_ref_filters
from spinta.datasets.keymaps.components import KeyMap
from spinta.datasets.utils import iterparams
from spinta.types.datatype import PrimaryKey, Array
from spinta.types.datatype import Ref
from spinta.typing import ObjectData
from spinta.ufuncs.querybuilder.components import QueryParams
from spinta.ufuncs.querybuilder.helpers import get_page_values
from spinta.ufuncs.helpers import merge_formulas
from spinta.ufuncs.resultbuilder.helpers import get_row_value, backend_result_builder_getter
from spinta.utils.nestedstruct import flat_dicts_to_nested, extract_list_property_names

log = logging.getLogger(__name__)


@commands.getall.register(Context, Model, Sql)
def getall(
    context: Context,
    model: Model,
    backend: Sql,
    *,
    query: Expr = None,
    params: QueryParams = None,
    extra_properties: dict[str, Property] = None,
    **kwargs
) -> Iterator[ObjectData]:
    conn = context.get(f'transaction.{backend.name}')
    builder = backend.query_builder_class(context)
    builder.update(model=model)
    # Merge user passed query with query set in manifest.
    query = merge_formulas(model.external.prepare, query)
    query = merge_formulas(query, get_enum_filters(context, model))
    query = merge_formulas(query, get_ref_filters(context, model))
    keymap: KeyMap = context.get(f'keymap.{model.keymap.name}')

    result_builder_getter = backend_result_builder_getter(context, backend)

    for model_params in iterparams(context, model, model.manifest):
        table = model.external.name.format(**model_params)
        table = backend.get_table(model, table)
        env = builder.init(backend, table, params)
        env.update(params=model_params)
        expr = env.resolve(query)
        where = env.execute(expr)
        qry = env.build(where)

        env_selected = env.selected
        list_keys = extract_list_property_names(model, env_selected.keys())
        is_page_enabled = env.page.page_.enabled

        for row in conn.execute(qry):
            res = {}

            for key, sel in env_selected.items():
                val = get_row_value(context, result_builder_getter, row, sel)
                if sel.prop:
                    if isinstance(sel.prop.dtype, PrimaryKey):
                        val = generate_pk_for_row(context, sel.prop.model, row, keymap, val)
                res[key] = val
            if is_page_enabled:
                res['_page'] = get_page_values(env, row)

            res['_type'] = model.model_type()
            res = flat_dicts_to_nested(res, list_keys=list_keys)
            res = commands.cast_backend_to_python(context, model, backend, res, extra_properties=extra_properties)
            yield res


@commands.getone.register(Context, Model, Sql)
def getone(
    context: Context,
    model: Model,
    backend: Sql,
    *,
    id_: str,
) -> ObjectData:
    keymap: KeyMap = context.get(f'keymap.{model.keymap.name}')
    _id = keymap.decode(model.name, id_)

    # preparing query for retrieving item by pk (single column or multi column)
    query = {}
    if isinstance(_id, list):
        pkeys = model.external.pkeys
        for index, pk in enumerate(pkeys):
            query[pk.name] = _id[index]
    else:
        pk = model.external.pkeys[0].name
        query[pk] = _id

    # building sqlalchemy query
    context.attach(f'transaction.{backend.name}', validate_and_return_begin, context, backend)
    conn = context.get(f'transaction.{backend.name}')
    table = model.external.name
    table = backend.get_table(model, table)

    qry = table.select()
    for column_name, column_value in query.items():
        id_column = table.c.get(column_name)
        qry = qry.where(id_column == column_value)

    # #executing query
    result = conn.execute(qry)
    row = result.fetchone()

    # preparing results
    data = {}

    for field in model.properties:
        if not field.startswith('_'):
            value = row[field]
            data[field] = value

    additional_data = {
        '_type': model.model_type(),
        '_id': id_
    }
    data.update(additional_data)
    data = flat_dicts_to_nested(data)
    return commands.cast_backend_to_python(context, model, backend, data)



import datetime
import pathlib
from contextlib import AsyncExitStack
from typing import Iterator

import aiofiles

from spinta import commands
from spinta.backends.postgresql.components import PostgreSQL
from spinta.cli.helpers.errors import cli_error
from spinta.cli.helpers.export.backends.postgresql.components import PostgresqlExportMetadata
from spinta.cli.helpers.export.backends.postgresql.helpers import split_data, generate_file_paths, \
    generate_export_metadata, prepare_changelog, generate_csv_headers, generate_table_data, extract_headers
from spinta.cli.helpers.export.components import CounterManager
from spinta.cli.helpers.export.helpers import create_data_items, prepare_data_without_checks, prepare_export_patch
from spinta.commands.write import prepare_data_for_write
from spinta.components import Context, Model, UrlParams, Action, DataSubItem
from spinta.core.enums import Access
from spinta.types.datatype import Array
from spinta.types.geometry.components import Geometry
from spinta.utils.aiotools import aiter, aenumerate
from spinta.utils.data import take


@commands.export_data.register(Context, Model, PostgreSQL)
async def export_data(
    context: Context,
    model: Model,
    backend: PostgreSQL,
    *,
    data: Iterator,
    path: str,
    counter: CounterManager,
    txn: str,
    access: Access,
    **kwargs
):
    params = UrlParams()
    params.action = Action.INSERT
    dstream = aiter(create_data_items(model, backend, data))
    dstream = prepare_data_without_checks(context, dstream)
    dstream = prepare_export_patch(context, dstream, access)
    dstream = prepare_data_for_write(context, dstream, params)
    data_headers, dstream = await extract_headers(context, model, backend, dstream, txn=txn)
    file_mapping, property_mapping = generate_file_paths(model, pathlib.Path(path))
    base_key = model.model_type()
    changes_key = f'{model.model_type()}.changes'

    async with AsyncExitStack() as stack:
        files = {
            key: await stack.enter_async_context(aiofiles.open(path, 'w+', newline=''))
            for key, path in file_mapping.items()
        }

        base_export_metadata = generate_export_metadata(
            model,
            base_key,
            files[base_key],
            fieldnames=data_headers,
            counter=counter
        )

        changes_export_metadata = generate_export_metadata(
            model,
            changes_key,
            files[changes_key],
            fieldnames=['_id', '_revision', '_txn', '_rid', 'datetime', 'action', 'data'],
            counter=counter
        )

        export_metadata: dict[str, PostgresqlExportMetadata] = {
            base_key: base_export_metadata,
            changes_key: changes_export_metadata,
            **{
                key: generate_export_metadata(
                    model,
                    key,
                    files[key],
                    fieldnames=generate_csv_headers(context, prop.dtype, backend),
                    counter=counter
                ) for key, prop in property_mapping.items() if key not in [base_key, changes_key]
            }
        }

        for meta in export_metadata.values():
            await meta.writeheader()

        async for i, item in aenumerate(dstream):
            created_time = datetime.datetime.now()
            data = commands.before_export(
                context,
                model,
                backend,
                data=item,
                txn=txn,
                created=created_time
            )
            normal_data, table_data = split_data(data)
            change_data = prepare_changelog(item.patch, i + 1, txn, created_time)
            await base_export_metadata.writerow(normal_data, counter)
            await changes_export_metadata.writerow(change_data, counter)

            for table in table_data:
                await export_metadata[table.name].write_table_data(table, normal_data, counter)

            counter.update_total(1)


@commands.before_export.register(Context, Model, PostgreSQL)
def before_export(
    context: Context,
    model: Model,
    backend: PostgreSQL,
    *,
    data: DataSubItem,
    txn: str = "",
    created: datetime.datetime = datetime.datetime.now()
):
    patch = take(['_id'], data.patch)
    patch['_revision'] = take('_revision', data.patch, data.saved)
    patch['_txn'] = txn
    patch['_created'] = created
    patch['_updated'] = None
    for prop in take(model.properties).values():
        if not prop.dtype.inherited:
            prop_data = data[prop.name]
            value = commands.before_export(
                context,
                prop.dtype,
                backend,
                data=prop_data,
            )
            patch.update(value)
    return patch


@commands.before_export.register(Context, Geometry, PostgreSQL)
def before_export(
    context: Context,
    dtype: Geometry,
    backend: PostgreSQL,
    *,
    data: DataSubItem
):
    value = data.patch
    if (
        dtype.srid is not None and
        value and
        "SRID" not in value
    ):
        value = f"SRID={dtype.srid};{value}"
    return take({dtype.prop.place: value})


@commands.before_export.register(Context, Array, PostgreSQL)
def before_export(
    context: Context,
    dtype: Array,
    backend: PostgreSQL,
    *,
    data: DataSubItem,
):
    if dtype.prop.list:
        return {}
    else:
        key = dtype.prop.place
        array_data = take({key: data.patch})
        table_data = generate_table_data(dtype, array_data[key])
        return {
            **array_data,
            table_data.name: table_data
        }


@commands.validate_export_output.register(Context, PostgreSQL, str)
def validate_export_output(context: Context, backend: PostgreSQL, output: str):
    commands.validate_export_output(
        context,
        backend,
        pathlib.Path(output)
    )


@commands.validate_export_output.register(Context, PostgreSQL, pathlib.Path)
def validate_export_output(context: Context, backend: PostgreSQL, output: pathlib.Path):
    if not output.exists():
        cli_error(
            f"{str(output)!r} directory does not exist."
        )



from collections.abc import Sequence
from typing import Union

import sqlalchemy as sa


def group_array(column: Union[sa.Column, Sequence[sa.Column]]):
    if isinstance(column, Sequence) and not isinstance(column, str):
        column = sa.sql.func.json_array(*column)
    # Using concat for mssql, because `json_arrayagg` is only available for Azure SQL version
    # https://learn.microsoft.com/en-us/sql/t-sql/functions/json-array-transact-sql?view=azuresqldb-current
    return sa.sql.func.concat('[', column, ']')



from typing import Any

from spinta.core.ufuncs import ufunc
from spinta.datasets.backends.sql.backends.sqlite.helpers import group_array
from spinta.datasets.backends.sql.backends.sqlite.ufuncs.query.components import SqliteQueryBuilder
from spinta.datasets.backends.sql.backends.helpers import nulls_last_asc, nulls_first_desc

import sqlalchemy as sa


@ufunc.resolver(SqliteQueryBuilder, object)
def _group_array(env: SqliteQueryBuilder, columns: Any):
    return group_array(columns)


@ufunc.resolver(SqliteQueryBuilder, sa.sql.expression.ColumnElement)
def asc(env, column):
    return nulls_last_asc(column)


@ufunc.resolver(SqliteQueryBuilder, sa.sql.expression.ColumnElement)
def desc(env, column):
    return nulls_first_desc(column)



from typing import Optional
from typing import Union
from urllib.parse import urlencode

from shapely.geometry.base import BaseGeometry

from pyproj import CRS, Transformer
from shapely.ops import transform

from shapely.geometry import Point

from spinta.exceptions import SRIDNotSetForGeometry
from spinta.types.geometry.components import Geometry
from spinta.types.geometry.constants import WGS84


def get_osm_link(value: BaseGeometry, srid: Optional[Union[int, Geometry]]) -> Optional[str]:
    if isinstance(srid, Geometry):
        if srid.srid is None:
            raise SRIDNotSetForGeometry(srid)
        srid = srid.srid

    if srid and srid != WGS84:
        transformer = Transformer.from_crs(
            crs_from=CRS(f'EPSG:{srid}'),
            crs_to=CRS(f'EPSG:{WGS84}'),
        )
        value = transform(transformer.transform, value)
    centroid = value.centroid

    # According to WGS84 (4326) `axis_info`
    # Axis order is:
    # x: Axis(name=Geodetic latitude, abbrev=Lat, direction=north, unit_auth_code=EPSG, unit_code=9122, unit_name=degree)
    # y: Axis(name=Geodetic longitude, abbrev=Lon, direction=east, unit_auth_code=EPSG, unit_code=9122, unit_name=degree)
    # Meaning x - lat, y - lon
    lat, lon = centroid.x, centroid.y
    params = urlencode({'mlat': lat, 'mlon': lon})

    return f'https://www.openstreetmap.org/?{params}#map=19/{lat}/{lon}'


def get_display_value(shape) -> Union[object, str]:
    if isinstance(shape, Point):
        value = shape.wkt
    else:
        # Shorten possibly long WKT values by reducing it to shape type.
        value = shape.geom_type.upper()
    return value



import sqlalchemy as sa


def nulls_last_asc(column: sa.Column):
    return sa.sql.expression.nullslast(column.asc())


def nulls_first_desc(column: sa.Column):
    return sa.sql.expression.nullsfirst(column.desc())



from copy import copy
from typing import Any, List, Union
from typing import Callable
from typing import Dict
from typing import Hashable
from typing import Iterator
from typing import Tuple
from typing import TypeVar

from spinta import commands
from spinta.cli.helpers.auth import require_auth
from spinta.cli.helpers.store import load_manifest
from spinta.components import Context, Property, Node
from spinta.components import Mode
from spinta.components import Model
from spinta.core.config import RawConfig, Path
from spinta.core.config import ResourceTuple
from spinta.core.config import parse_resource_args
from spinta.core.context import configure_context, create_context
from spinta.datasets.components import Dataset, Resource, ExternalBackend
from spinta.datasets.inspect.components import PriorityKey
from spinta.exceptions import InvalidResourceSource
from spinta.manifests.components import Manifest
from spinta.manifests.components import ManifestPath
from spinta.manifests.helpers import init_manifest
from spinta.utils.naming import Deduplicator
from spinta.utils.schema import NA


def create_manifest_from_inspect(
    context: Context = None,
    manifest: Union[str, ManifestPath] = None,
    resources: Tuple = None,
    backend: str = None,
    formula: str = None,
    dataset: str = None,
    auth: str = None,
    priority: str = "manifest",
    only_url: bool = False,
):
    # Reset Context, to not have any outside influence from given
    if context is None:
        rc = RawConfig()
        rc.read([Path('spinta', 'spinta.config:CONFIG')])
        context = create_context(name="inspect", rc=rc)

    has_manifest_priority = priority == 'manifest'
    if resources:
        resources = parse_resource_args(*resources, formula)

    context = configure_context(
        context,
        [manifest] if manifest else None,
        mode=Mode.external,
        backend=backend
    )
    with context:
        require_auth(context, auth)
        store = load_manifest(context, ensure_config_dir=True, full_load=True)
        old = store.manifest
        manifest = Manifest()
        init_manifest(context, manifest, 'inspect')
        commands.merge(context, manifest, manifest, old, has_manifest_priority)

        if not resources:
            resources = []
            for ds in commands.get_datasets(context, old).values():
                for resource in ds.resources.values():
                    external = resource.external
                    if external == '' and resource.backend:
                        external = resource.backend.config['dsn']
                    if not any(res.external == external for res in resources):
                        if only_url and not ("http://" in external or "https://" in external):
                            raise InvalidResourceSource(source=external)
                        resources.append(ResourceTuple(type=resource.type, external=external, prepare=resource.prepare))

        if resources:
            for resource in resources:
                _merge(context, manifest, manifest, resource, has_manifest_priority, dataset)

        # Sort models for render
        sorted_models = {}
        for key, model in commands.get_models(context, manifest).items():
            if key not in sorted_models.keys():
                if model.external and model.external.resource:
                    resource = model.external.resource
                    for resource_key, resource_model in resource.models.items():
                        if resource_key not in sorted_models.keys():
                            sorted_models[resource_key] = resource_model
                else:
                    sorted_models[key] = model
        commands.set_models(context, manifest, sorted_models)
    return context, manifest


def _merge(context: Context, manifest: Manifest, old: Manifest, resource: ResourceTuple, has_manifest_priority: bool, dataset: str = None):
    manifest_ = commands.backend_to_manifest_type(context, resource.type)
    path = ManifestPath(type=manifest_.type, path=resource.external, prepare=resource.prepare)
    context = configure_context(context, [path], mode=Mode.external, dataset=dataset)
    store = load_manifest(context, full_load=True)
    new = store.manifest
    commands.merge(context, manifest, old, new, has_manifest_priority)


def _filter_models_for_dataset(
    context: Context,
    manifest: Manifest,
    dataset: Dataset
) -> List[Model]:
    models = []
    for model in commands.get_models(context, manifest).values():
        if model.external:
            if model.external.dataset is dataset:
                models.append(model)
    return models


def _merge_model_properties(
    context: Context,
    manifest: Manifest,
    old: Model,
    new: Model,
    has_manifest_priority: bool
):
    properties = zipitems(
        old.properties.values(),
        new.properties.values(),
        _property_key,
    )
    deduplicator = Deduplicator("_{}")
    for prop in properties:
        for o, n in prop:
            if n:
                n = copy(n)
                n.model = new
            if o:
                deduplicator(o.basename)
                o.model = old
            if n and not o:
                name = deduplicator(n.basename)
                n.model = old
                n.name = name
                n.place = name
            commands.merge(context, manifest, o, n, has_manifest_priority)


def _merge_prefixes(
    context: Context,
    manifest: Manifest,
    old: Dataset,
    new: Dataset
):
    prefixes = zipitems(
        old.prefixes.values(),
        new.prefixes.values(),
        _name_key,
    )
    for prefix in prefixes:
        for o, n in prefix:
            commands.merge(context, manifest, o, n)


def _merge_resources(
    context: Context,
    manifest: Manifest,
    old: Dataset,
    new: Dataset
):
    old_resources = [] if old == NA else old.resources.values()
    new_resources = [] if new == NA else new.resources.values()
    resources = zipitems(
        old_resources,
        new_resources,
        _resource_key,
    )
    deduplicator = Deduplicator("{}")
    for res in resources:
        for o, n in res:
            if o:
                deduplicator(o.name)
            elif n:
                n.name = deduplicator(n.name)
            commands.merge(context, manifest, o, n)


TItem = TypeVar('TItem')


def zipitems(
    a: TItem,
    b: TItem,
    key: Callable[[TItem], Hashable],
) -> Iterator[List[Tuple[TItem, TItem]]]:

    res: Dict[
        Hashable,  # key
        List[
            List[
                Any,   # a
                Any,   # b
            ]
        ]
        ,
    ] = {}

    # Map first values
    for value in a:
        value_key = key(value)
        new_value = [value, NA]

        if isinstance(value_key, Tuple):
            for existing_key in res.keys():
                if set(value_key) == set(existing_key):
                    value_key = existing_key
                    break
        elif isinstance(value_key, PriorityKey):
            for existing_key in res.keys():
                if value_key == existing_key:
                    value_key = existing_key
                    break

        res.setdefault(value_key, []).append(new_value)

    # Map second value
    for value in b:
        value_key = key(value)
        new_value = [NA, value]

        mapped_keys = []
        if isinstance(value_key, Tuple):
            value_key_set = set(value_key)
            mapped_keys = [existing_key for existing_key in res if value_key_set.issubset(set(existing_key))]
        elif isinstance(value_key, PriorityKey):
            # Only get the first match
            for existing_key in res:
                if existing_key == value_key:
                    mapped_keys = [existing_key]
                    break

        if not mapped_keys:
            mapped_keys = [value_key]

        for mapped_key in mapped_keys:
            if mapped_key not in res or not res[mapped_key]:
                res[mapped_key] = [new_value]
                continue

            extension_list = []
            for existing_value in res[mapped_key]:
                if existing_value[1] is NA:
                    existing_value[1] = value
                else:
                    extension_list.append([existing_value[0], value])
            res[mapped_key] += extension_list

    yield from res.values()


def coalesce(*args: Any) -> Any:
    for arg in args:
        if arg:
            return arg
    return arg


def _name_key(node: Node) -> str:
    return node.name


def _dataset_key(dataset: Dataset) -> PriorityKey:
    key = PriorityKey()
    if dataset.id:
        key.id = dataset.id
    if dataset.given.name:
        key.name = dataset.given.name
    key.source = _dataset_resource_source_key(dataset)
    return key


def _dataset_resource_source_key(dataset: Dataset) -> Tuple:
    keys = []
    for resource in dataset.resources.values():
        keys.append(_resource_source_key(resource))
    return tuple(keys)


def _property_key(prop: Property) -> PriorityKey:
    key = PriorityKey()
    if prop.id:
        key.id = prop.id
    if prop.given.name:
        key.name = prop.given.name
    key.source = _property_source_key(prop)
    return key


def _property_source_key(prop: Property) -> str:
    if prop.external and prop.external.name:
        return prop.external.name
    else:
        return prop.name


def _resource_key(resource: Resource) -> PriorityKey:
    key = PriorityKey()
    if resource.id:
        key.id = resource.id
    if resource.given.name:
        key.name = resource.given.name
    key.source = _resource_source_key(resource)
    return key


def _resource_source_key(resource: Resource) -> str:
    result = resource.name
    manifest = resource.dataset.manifest
    if resource.external:
        result = resource.external
    else:
        if manifest.backends:
            if resource.backend.name in manifest.backends:
                result = manifest.backends[resource.backend.name].config['dsn']
        elif manifest.store.backends:
            if resource.backend.name in manifest.store.backends:
                result = manifest.store.backends[resource.backend.name].config['dsn']
    if "@" in result:
        result = result.split("@")[1]
    return result


def _model_key(model: Model) -> PriorityKey:
    key = PriorityKey()
    if model.id:
        key.id = model.id
    if model.given.name:
        key.name = model.given.name
    key.source = _model_source_key(model)
    return key


def _model_source_key(model: Model) -> str:
    if model.external and model.external.name:
        return f'{_resource_source_key(model.external.resource)}/{model.external.name}'
    else:
        return model.name


def _backend_dsn(backend: ExternalBackend) -> str:
    dsn = backend.config['dsn']
    if "@" in dsn:
        dsn = dsn.split("@")[1]
    return dsn






from spinta.components import UrlParams, ParamsPage


def disable_params_pagination(params: UrlParams):
    if params.page is not None:
        params.page.is_enabled = False
    else:
        params.page = ParamsPage(is_enabled=False)



from copy import copy

from spinta import commands
from spinta.components import Context, Property
from spinta.exceptions import PartialTypeNotFound, ParentNodeNotFound
from spinta.types.datatype import Partial, Ref


def get_ref_value(context: Context, prop: Property):
    parent = prop.parent
    if isinstance(parent, Property) and isinstance(parent.dtype, Ref):
        parent_parent = parent.parent
        if isinstance(parent_parent, Property) and isinstance(parent_parent.dtype, Ref):
            model = parent_parent.dtype.model
            first_level = model.properties[parent.name]
            if prop.name in first_level.dtype.properties:
                return first_level.dtype.properties[prop.name]
        else:
            model = parent.dtype.model
            if prop.name in model.properties:
                return model.properties[prop.name]


@commands.link.register(Context, Partial)
def link(context: Context, dtype: Partial):
    prop = dtype.prop
    parent = prop.parent
    if parent and isinstance(parent, Property):
        if isinstance(parent.dtype, Ref):
            props = dtype.properties
            result = get_ref_value(context, prop)
            prop.dtype = copy(result.dtype)
            prop.dtype.properties = props
            prop.dtype.inherited = True
            prop.given.explicit = False
            prop.given.name = ''
            prop.dtype.prop = prop
            for partial_prop in props.values():
                commands.link(context, partial_prop)
        else:
            raise PartialTypeNotFound(dtype)
    else:
        raise ParentNodeNotFound(
            dtype,
            model_name=prop.model.name,
            property_names=','.join([f'{prop.name}.{i}' for i in dtype.properties.keys()]),
            missing_property_name=prop.name,
        )




from typing import Any

from spinta.components import Property
from spinta.core.ufuncs import Env


class ResultBuilder(Env):
    this: Any  # A selected value.
    prop: Property  # Model property.
    data: dict  # Whole row from database.

    def init(self, this: Any, prop: Property, data: dict):
        return self(
            this=this,
            prop=prop,
            data=data,
        )



from functools import reduce
from typing import Dict, Any, Tuple, List

from spinta.auth import authorized
from spinta.components import Property, Action
from spinta.core.ufuncs import Expr, ufunc, Bind, Unresolved, GetAttr
from spinta.datasets.backends.dataframe.ufuncs.query.components import DaskDataFrameQueryBuilder, DaskSelected as Selected
from spinta.exceptions import PropertyNotFound, NotImplementedFeature, SourceCannotBeList
from spinta.types.datatype import DataType, PrimaryKey, Ref, Binary
from spinta.types.text.components import Text
from spinta.ufuncs.components import ForeignProperty
from spinta.utils.data import take
from spinta.utils.schema import NA
import base64 as b64


@ufunc.resolver(DaskDataFrameQueryBuilder, Expr, name='and')
def and_(env, expr):
    args, kwargs = expr.resolve(env)
    args = [a for a in args if a is not None]
    return env.call('and', args)


@ufunc.resolver(DaskDataFrameQueryBuilder, list, name='and')
def and_(env, args):
    if len(args) > 1:
        return reduce(lambda x, y: x & y, args)
    elif args:
        return args[0]


@ufunc.resolver(DaskDataFrameQueryBuilder, Expr, name='or')
def or_(env, expr):
    args, kwargs = expr.resolve(env)
    args = [a for a in args if a is not None]
    return env.call('or', args)


@ufunc.resolver(DaskDataFrameQueryBuilder, list, name='or')
def or_(env, args):
    if len(args) > 1:
        return reduce(lambda x, y: x | y, args)
    elif args:
        return args[0]


@ufunc.resolver(DaskDataFrameQueryBuilder)
def distinct(env: DaskDataFrameQueryBuilder):
    if env.model and env.model.external and not env.model.external.unknown_primary_key:
        extracted_columns = [prop.external.name for prop in env.model.external.pkeys if prop.external]
        env.dataframe = env.dataframe.drop_duplicates(extracted_columns)
    else:
        env.dataframe = env.dataframe.drop_duplicates()


@ufunc.resolver(DaskDataFrameQueryBuilder, int)
def limit(env: DaskDataFrameQueryBuilder, n: int):
    env.limit = n


@ufunc.resolver(DaskDataFrameQueryBuilder, int)
def offset(env: DaskDataFrameQueryBuilder, n: int):
    env.offset = n


@ufunc.resolver(DaskDataFrameQueryBuilder, GetAttr)
def _resolve_property(
    env: DaskDataFrameQueryBuilder,
    attr: GetAttr
):
    return env.call('_resolve_property', attr.obj)


@ufunc.resolver(DaskDataFrameQueryBuilder, Bind)
def _resolve_property(
    env: DaskDataFrameQueryBuilder,
    bind: Bind
):
    return env.call('_resolve_property', bind.name)


@ufunc.resolver(DaskDataFrameQueryBuilder, str)
def _resolve_property(
    env: DaskDataFrameQueryBuilder,
    prop: str
):
    if prop in env.model.flatprops:
        return env.model.flatprops.get(prop)

    raise PropertyNotFound(env.model, property=prop)


@ufunc.resolver(DaskDataFrameQueryBuilder, Property)
def _resolve_property(
    env: DaskDataFrameQueryBuilder,
    prop: Property
):
    return prop


@ufunc.resolver(DaskDataFrameQueryBuilder, object)
def _resolve_unresolved(env: DaskDataFrameQueryBuilder, value: Any) -> Any:
    if isinstance(value, Unresolved):
        raise ValueError(f"Unresolved value {value!r}.")
    else:
        return value


@ufunc.resolver(DaskDataFrameQueryBuilder, Bind)
def _resolve_unresolved(env: DaskDataFrameQueryBuilder, field: Bind) -> str:
    prop = env.model.flatprops.get(field.name)
    if prop:
        return prop.external.name
    else:
        raise PropertyNotFound(env.model, property=field.name)


@ufunc.resolver(DaskDataFrameQueryBuilder)
def count(env: DaskDataFrameQueryBuilder):
    return len(env.dataframe.index)


@ufunc.resolver(DaskDataFrameQueryBuilder, Expr)
def select(env: DaskDataFrameQueryBuilder, expr: Expr):
    keys = [str(k) for k in expr.args]
    args, kwargs = expr.resolve(env)
    args = list(zip(keys, args)) + list(kwargs.items())

    if env.selected is not None:
        raise RuntimeError("`select` was already called.")

    env.selected = {}
    if args:
        for key, arg in args:
            env.selected[key] = env.call('select', arg)
    else:
        for prop in take(['_id', all], env.model.properties).values():
            if authorized(env.context, prop, Action.GETALL):
                env.selected[prop.place] = env.call('select', prop)


@ufunc.resolver(DaskDataFrameQueryBuilder, Bind)
def select(env: DaskDataFrameQueryBuilder, item: Bind, *, nested: bool = False):
    prop = _get_property_for_select(env, item.name, nested=nested)
    return env.call('select', prop)


def _get_property_for_select(
    env: DaskDataFrameQueryBuilder,
    name: str,
    *,
    nested: bool = False,
) -> Property:
    # TODO: `name` can refer to (in specified order):
    #       - var - a defined variable
    #       - param - a parameter if parametrization is used
    #       - item - an item of a dict or list
    #       - prop - a property
    #       Currently only `prop` is resolved.
    prop = env.call('_resolve_property', name)
    if (
        # Check authorization only for top level properties in select list.
        # XXX: Not sure if nested is the right property to user, probably better
        #      option is to check if this call comes from a prepare context. But
        #      then how prepare context should be defined? Probably resolvers
        #      should be called with a different env class?
        #      tag:resolving_private_properties_in_prepare_context
        nested or
        authorized(env.context, prop, Action.SEARCH)
    ):
        return prop
    else:
        raise PropertyNotFound(env.model, property=name)


@ufunc.resolver(DaskDataFrameQueryBuilder, Property)
def select(env: DaskDataFrameQueryBuilder, prop: Property) -> Selected:
    if prop.place not in env.resolved:
        if isinstance(prop.external, list):
            raise SourceCannotBeList(prop)
        if prop.external.prepare is not NA:
            # If `prepare` formula is given, evaluate formula.
            if isinstance(prop.external.prepare, Expr):
                result = env(this=prop).resolve(prop.external.prepare)
            else:
                result = prop.external.prepare
            # XXX: Maybe interpretation of prepare formula should be done under
            #      a different Env? This way, select resolvers would know when
            #      properties are resolved under a formula context and for
            #      example would be able to decide to not check permissions on
            #      properties.
            #      tag:resolving_private_properties_in_prepare_context
            result = env.call('select', prop.dtype, result)
        elif prop.external and prop.external.name:
            # If prepare is not given, then take value from `source`.
            result = env.call('select', prop.dtype)
        elif prop.is_reserved():
            # Reserved properties never have external source.
            result = env.call('select', prop.dtype)
        elif not prop.dtype.requires_source:
            # Some property types do not require source (object, text, etc)
            result = env.call('select', prop.dtype)
        else:
            # If `source` is not given, return None.
            result = Selected(prop=prop, prep=None)
        assert isinstance(result, Selected), prop
        env.resolved[prop.place] = result
    return env.resolved[prop.place]


@ufunc.resolver(DaskDataFrameQueryBuilder, DataType)
def select(env: DaskDataFrameQueryBuilder, dtype: DataType) -> Selected:
    return Selected(
        item=dtype.prop.external.name,
        prop=dtype.prop,
    )


@ufunc.resolver(DaskDataFrameQueryBuilder, Text)
def select(env: DaskDataFrameQueryBuilder, dtype: Text) -> Selected:
    prep = {}
    for lang, prop in dtype.langs.items():
        prep[lang] = env.call('select', prop)
    return Selected(prop=dtype.prop, prep=prep)


@ufunc.resolver(DaskDataFrameQueryBuilder, DataType, object)
def select(env: DaskDataFrameQueryBuilder, dtype: DataType, prep: Any) -> Selected:
    if isinstance(prep, str):
        # XXX: Backwards compatibility thing.
        #      str values are interpreted as Bind values and Bind values are
        #      assumed to be properties. So here we skip
        #      `env.call('select', prep)` and return `prep` as is.
        #      This should be eventually removed, once backwards compatibility
        #      for resolving strings as properties is removed.
        return Selected(prop=dtype.prop, prep=prep)
    elif isinstance(prep, Expr):
        # If `prepare` expression returns another expression, then this means,
        # it must be processed on values returned by query.
        prop = dtype.prop
        if prop.external and prop.external.name:
            sel = env.call('select', dtype)
            return Selected(item=sel.item, prop=sel.prop, prep=prep)
        else:
            return Selected(item=None, prop=prop, prep=prep)
    else:
        if prep is not None:
            result = env.call('select', prep)
        else:
            result = None
        if isinstance(result, Selected):
            return result
        return Selected(prop=dtype.prop, prep=result)


@ufunc.resolver(DaskDataFrameQueryBuilder, PrimaryKey)
def select(
    env: DaskDataFrameQueryBuilder,
    dtype: PrimaryKey,
) -> Selected:
    model = dtype.prop.model
    pkeys = model.external.pkeys

    if not pkeys:
        # If primary key is not specified use all properties to uniquely
        # identify row.
        pkeys = take(model.properties).values()

    if len(pkeys) == 1:
        prop = pkeys[0]
        result = env.call('select', prop)
    else:
        result = [
            env.call('select', prop)
            for prop in pkeys
        ]
    return Selected(prop=dtype.prop, prep=result)


@ufunc.resolver(DaskDataFrameQueryBuilder, Selected)
def select(env: DaskDataFrameQueryBuilder, selected: Selected):
    return selected


@ufunc.resolver(DaskDataFrameQueryBuilder, Ref, object)
def select(env: DaskDataFrameQueryBuilder, dtype: Ref, prep: Any) -> Selected:
    fpr = ForeignProperty(None, dtype, dtype.model.properties['_id'].dtype)
    return Selected(
        prop=dtype.prop,
        prep=env.call('select', fpr, fpr.right.prop),
    )


@ufunc.resolver(DaskDataFrameQueryBuilder, GetAttr)
def select(env: DaskDataFrameQueryBuilder, attr: GetAttr) -> Selected:
    """For things like select(foo.bar.baz)."""

    fpr: ForeignProperty = env.call('_resolve_getattr', attr)
    raise NotImplementedFeature(fpr.left.prop.model, feature="Ability to use foreign properties")
    return Selected(
        prop=fpr.right.prop,
        prep=env.call('select', fpr, fpr.right.prop),
    )


@ufunc.resolver(DaskDataFrameQueryBuilder, ForeignProperty)
def select(
    env: DaskDataFrameQueryBuilder,
    fpr: ForeignProperty,
) -> Selected:
    return env.call('select', fpr, fpr.right.prop)


@ufunc.resolver(DaskDataFrameQueryBuilder, ForeignProperty, Property)
def select(
    env: DaskDataFrameQueryBuilder,
    fpr: ForeignProperty,
    prop: Property,
) -> Selected:
    resolved_key = fpr / prop
    if resolved_key not in env.resolved:
        if isinstance(prop.external, list):
            raise ValueError("Source can't be a list, use prepare instead.")
        if prop.external.prepare is not NA:

            if isinstance(prop.external.prepare, Expr):
                result = env(this=prop).resolve(prop.external.prepare)
            else:
                result = prop.external.prepare
            result = env.call('select', fpr, prop.dtype, result)
        else:
            result = env.call('select', fpr, prop.dtype)
        assert isinstance(result, Selected), prop
        env.resolved[resolved_key] = result
    return env.resolved[resolved_key]


@ufunc.resolver(DaskDataFrameQueryBuilder, ForeignProperty, DataType)
def select(
    env: DaskDataFrameQueryBuilder,
    fpr: ForeignProperty,
    dtype: DataType,
) -> Selected:
    # TODO need join for this to work
    return Selected(
        item=dtype.prop.name,
        prop=dtype.prop,
    )


@ufunc.resolver(DaskDataFrameQueryBuilder, ForeignProperty, DataType, object)
def select(
    env: DaskDataFrameQueryBuilder,
    fpr: ForeignProperty,
    dtype: DataType,
    prep: Any,
) -> Selected:
    result = env.call('select', fpr, prep)
    return Selected(prop=dtype.prop, prep=result)


@ufunc.resolver(DaskDataFrameQueryBuilder, ForeignProperty, tuple)
def select(
    env: DaskDataFrameQueryBuilder,
    fpr: ForeignProperty,
    prep: Tuple[Any],
) -> Tuple[Any]:
    return tuple(env.call('select', fpr, v) for v in prep)


@ufunc.resolver(DaskDataFrameQueryBuilder, ForeignProperty, Bind)
def select(env: DaskDataFrameQueryBuilder, fpr: ForeignProperty, item: Bind):
    model = fpr.right.prop.model
    prop = env.call('_resolve_property', item)
    if authorized(env.context, prop, Action.SEARCH):
        return env.call('select', fpr, prop)
    else:
        raise PropertyNotFound(model, property=item.name)


@ufunc.resolver(DaskDataFrameQueryBuilder, ForeignProperty, PrimaryKey)
def select(
    env: DaskDataFrameQueryBuilder,
    fpr: ForeignProperty,
    dtype: PrimaryKey,
) -> Selected:
    model = dtype.prop.model
    pkeys = model.external.pkeys

    if not pkeys:
        raise RuntimeError(
            f"Can't join {dtype.prop} on right table without primary key."
        )

    if len(pkeys) == 1:
        prop = pkeys[0]
        result = env.call('select', fpr, prop)
    else:
        result = [
            env.call('select', fpr, prop)
            for prop in pkeys
        ]
    return Selected(prop=dtype.prop, prep=result)


@ufunc.resolver(DaskDataFrameQueryBuilder, list)
def select(
    env: DaskDataFrameQueryBuilder,
    prep: List[Any],
) -> List[Any]:
    return [env.call('select', v) for v in prep]


@ufunc.resolver(DaskDataFrameQueryBuilder, tuple)
def select(
    env: DaskDataFrameQueryBuilder,
    prep: Tuple[Any],
) -> Tuple[Any]:
    return tuple(env.call('select', v) for v in prep)


@ufunc.resolver(DaskDataFrameQueryBuilder, dict)
def select(
    env: DaskDataFrameQueryBuilder,
    prep: Dict[str, Any],
) -> Dict[str, Any]:
    return {k: env.call('select', v) for k, v in prep.items()}


@ufunc.resolver(DaskDataFrameQueryBuilder)
def base64(env: DaskDataFrameQueryBuilder) -> bytes:
    return env.call('base64', env.this)


@ufunc.resolver(DaskDataFrameQueryBuilder, Property)
def base64(env: DaskDataFrameQueryBuilder, prop: Property) -> bytes:
    return env.call('base64', prop.dtype)


@ufunc.resolver(DaskDataFrameQueryBuilder, Binary)
def base64(env: DaskDataFrameQueryBuilder, dtype: Binary) -> Selected:
    item = f'base64({dtype.prop.external.name})'
    env.dataframe[item] = env.dataframe[dtype.prop.external.name].str.encode('ascii').map(b64.decodebytes, meta=(dtype.prop.external.name, bytes))
    return Selected(
        item=item,
        prop=dtype.prop
    )


COMPARE = [
    'eq',
    'ne',
    'lt',
    'le',
    'gt',
    'ge',
    'startswith',
    'contains',
]


@ufunc.resolver(DaskDataFrameQueryBuilder, Bind, object, names=COMPARE)
def compare(env, op, field, value):
    prop = env.call('_resolve_property', field)
    return env.call(op, prop.dtype, value)


@ufunc.resolver(DaskDataFrameQueryBuilder, DataType, object, name="eq")
def eq_(env: DaskDataFrameQueryBuilder, dtype: DataType, obj: object):
    name = dtype.prop.external.name
    return env.dataframe[name] == str(obj)





from __future__ import annotations

import dataclasses
from typing import List, Union, Any

import pymongo

from spinta.components import Property
from spinta.core.ufuncs import Expr
from spinta.exceptions import UnknownMethod
from spinta.types.datatype import DataType
from spinta.ufuncs.querybuilder.components import QueryBuilder, QueryPage
from spinta.ufuncs.querybuilder.helpers import merge_with_page_selected_list, merge_with_page_sort, \
    merge_with_page_limit
from spinta.backends.mongo.components import Mongo


class MongoQueryBuilder(QueryBuilder):

    def init(self, backend: Mongo, table: pymongo.collection.Collection):
        return self(
            backend=backend,
            table=table,
            select=None,
            sort=[],
            limit=None,
            offset=None,
            page=QueryPage()
        )

    def build(self, where: list):
        if self.select is None:
            self.call('select', Expr('select'))

        select = []
        merged_selected = merge_with_page_selected_list(list(self.select.values()), self.page)
        merged_sorted = merge_with_page_sort(self.sort, self.page)
        merged_limit = merge_with_page_limit(self.limit, self.page)
        for sel in merged_selected:
            if isinstance(sel.item, list):
                select += sel.item
            else:
                select.append(sel.item)

        select = {k: 1 for k in select}
        select['_id'] = 0
        select['__id'] = 1
        select['_revision'] = 1

        where = where or {}

        cursor = self.table.find(where, select)

        if merged_sorted:
            cursor = cursor.sort(merged_sorted)

        if merged_limit is not None:
            cursor = cursor.limit(merged_limit)

        if self.offset is not None:
            cursor = cursor.skip(self.offset)

        return cursor

    def default_resolver(self, expr, *args, **kwargs):
        raise UnknownMethod(expr=str(expr(*args, **kwargs)), name=expr.name)


class ForeignProperty:

    def __init__(self, fpr: ForeignProperty, left: Property, right: Property):
        if fpr is None:
            self.name = left.place
            self.chain = [self]
        else:
            self.name += '->' + left.place
            self.chain = fpr.chain + [self]

        self.left = left
        self.right = right

    def __repr__(self):
        return f'<{self.name}->{self.right.name}:{self.right.dtype.name}>'


class Func:
    pass


@dataclasses.dataclass
class Lower(Func):
    dtype: DataType = None


@dataclasses.dataclass
class Negative(Func):
    arg: Any


@dataclasses.dataclass
class Positive(Func):
    arg: Any


@dataclasses.dataclass
class Recurse(Func):
    args: List[Union[DataType, Func]] = None



from typing import Any


class PriorityKey:
    id: Any = None
    name: str = None
    source: Any = None

    def __init__(self, _id=None, name=None, source=None):
        self.id = _id
        self.name = name
        self.source = source

    def __eq__(self, other):
        if isinstance(other, PriorityKey):
            if self.id and other.id:
                if self.id == other.id:
                    return True
            if self.name and other.name:
                if self.name == other.name:
                    return True
            if self.source and other.source:
                if isinstance(other.source, tuple) or isinstance(self.source, tuple):
                    if set(self.source).issubset(other.source) or set(other.source).issubset(self.source):
                        return True
                else:
                    if self.source == other.source:
                        return True
        return False

    def __str__(self):
        return f"PriorityKey(id: {self.id}, name: {self.name}, source: {self.source})"

    def __hash__(self):
        return hash(str(self))



from collections.abc import Sequence
from typing import Union

import sqlalchemy as sa
import geoalchemy2


def flip_geometry_postgis(column: sa.Column):
    return geoalchemy2.functions.ST_FlipCoordinates(column)


def group_array(column: Union[sa.Column, Sequence[sa.Column]]):
    if isinstance(column, Sequence) and not isinstance(column, str):
        column = sa.sql.func.jsonb_build_array(*column)
    return sa.sql.func.jsonb_agg(column)



from __future__ import annotations

import dataclasses
import datetime
import re
from typing import Any

import pymongo

from spinta import exceptions
from spinta.auth import authorized
from spinta.backends.mongo.ufuncs.components import MongoQueryBuilder, Recurse, Lower, Negative, Positive
from spinta.components import Action, Model, Property, Page
from spinta.core.ufuncs import Bind
from spinta.core.ufuncs import Expr
from spinta.core.ufuncs import ufunc
from spinta.exceptions import EmptyStringSearch
from spinta.exceptions import FieldNotInResource
from spinta.types.datatype import Array
from spinta.types.datatype import DataType
from spinta.types.datatype import Date
from spinta.types.datatype import DateTime
from spinta.types.datatype import Integer
from spinta.types.datatype import Number
from spinta.types.datatype import Object
from spinta.types.datatype import PrimaryKey
from spinta.types.datatype import String
from spinta.ufuncs.querybuilder.ufuncs import Star
from spinta.utils.data import take


@ufunc.resolver(MongoQueryBuilder, str)
def op(env, arg: str):
    if arg == '*':
        return Star()
    else:
        raise NotImplementedError


@ufunc.resolver(MongoQueryBuilder, Bind, Bind, name='getattr')
def getattr_(env, field, attr):
    if field.name in env.model.properties:
        prop = env.model.properties[field.name]
    else:
        raise FieldNotInResource(env.model, property=field.anem)
    return env.call('getattr', prop.dtype, attr)


@ufunc.resolver(MongoQueryBuilder, Object, Bind)
def getattr(env, dtype, attr):
    if attr.name in dtype.properties:
        return dtype.properties[attr.name].dtype
    else:
        raise FieldNotInResource(dtype, property=attr.name)


@ufunc.resolver(MongoQueryBuilder, Array, Bind)
def getattr(env, dtype, attr):
    return env.call('getattr', dtype.items.dtype, attr)


@ufunc.resolver(MongoQueryBuilder, Expr)
def select(env, expr):
    keys = [str(k) for k in expr.args]
    args, kwargs = expr.resolve(env)
    args = list(zip(keys, args)) + list(kwargs.items())

    env.select = {}

    if args:
        for key, arg in args:
            selected = env.call('select', arg)
            if selected is not None:
                env.select[key] = selected
    else:
        env.call('select', Star())

    if not (len(args) == 1 and args[0][0] == '_page'):
        assert env.select, args


@ufunc.resolver(MongoQueryBuilder, Star)
def select(env, arg: Star) -> None:
    for prop in take(env.model.properties).values():
        # if authorized(env.context, prop, (Action.GETALL, Action.SEARCH)):
        # TODO: This line above should come from a getall(request),
        #       because getall can be used internally for example for
        #       writes.
        env.select[prop.place] = env.call('select', prop.dtype)


@ufunc.resolver(MongoQueryBuilder, Bind)
def select(env, field):
    if field.name == '_page':
        return None

    prop = env.resolve_property(field)
    if authorized(env.context, prop, Action.SEARCH):
        return env.call('select', prop.dtype)
    else:
        raise FieldNotInResource(env.model, property=field.name)


@dataclasses.dataclass
class Selected:
    item: Any
    prop: Property = None


@ufunc.resolver(MongoQueryBuilder, DataType)
def select(env, dtype):
    table = env.backend.get_table(env.model)
    column = env.backend.get_column(table, dtype.prop, select=True)
    return Selected(column, dtype.prop)


@ufunc.resolver(MongoQueryBuilder, Page)
def select(env, page):
    return_selected = []
    for item in page.by.values():
        selected = env.call('select', item.prop.dtype)
        return_selected.append(selected)
    return return_selected


@ufunc.resolver(MongoQueryBuilder, int)
def limit(env, n):
    env.limit = n


@ufunc.resolver(MongoQueryBuilder, int)
def offset(env, n):
    env.offset = n


@ufunc.resolver(MongoQueryBuilder, Expr, name='and')
def and_(env, expr):
    args, kwargs = expr.resolve(env)
    args = [a for a in args if a is not None]
    return env.call('and', args)


@ufunc.resolver(MongoQueryBuilder, list, name='and')
def and_(env, args):
    if len(args) > 1:
        return {'$and': args}
    elif args:
        return args[0]


@ufunc.resolver(MongoQueryBuilder, Expr, name='or')
def or_(env, expr):
    args, kwargs = expr.resolve(env)
    args = [a for a in args if a is not None]
    return env.call('or', args)


@ufunc.resolver(MongoQueryBuilder, list, name='or')
def or_(env, args):
    if len(args) > 1:
        return {'$or': args}
    elif args:
        return args[0]


COMPARE = [
    'eq',
    'ne',
    'lt',
    'le',
    'gt',
    'ge',
    'startswith',
    'contains',
]


@ufunc.resolver(MongoQueryBuilder, DataType, object, names=[
    'eq', 'lt', 'le', 'gt', 'ge', 'contains', 'startswith',
])
def compare(env, op, dtype, value):
    raise exceptions.InvalidValue(dtype, op=op, value=type(value))


@ufunc.resolver(MongoQueryBuilder, PrimaryKey, (object, type(None)), names=[
    'eq', 'lt', 'le', 'gt', 'ge',
])
def compare(env, op, dtype, value):
    return _mongo_compare(op, '__id', value)


@ufunc.resolver(MongoQueryBuilder, DataType, type(None))
def eq(env, dtype, value):
    return {dtype.prop.place: None}


@ufunc.resolver(MongoQueryBuilder, String, (str, re.Pattern))
def eq(env, dtype, value):
    return _mongo_compare('eq', dtype.prop.place, value)


@ufunc.resolver(MongoQueryBuilder, (Integer, Number), (int, float), names=[
    'eq', 'lt', 'le', 'gt', 'ge',
])
def compare(env, op, dtype, value):
    return _mongo_compare(op, dtype.prop.place, value)


@ufunc.resolver(MongoQueryBuilder, DateTime, str, names=[
    'eq', 'lt', 'le', 'gt', 'ge',
])
def compare(env, op, dtype, value):
    value = datetime.datetime.fromisoformat(value)
    return _mongo_compare(op, dtype.prop.place, value)


@ufunc.resolver(MongoQueryBuilder, Date, str, names=[
    'eq', 'lt', 'le', 'gt', 'ge',
])
def compare(env, op, dtype, value):
    value = datetime.date.fromisoformat(value)
    value = datetime.datetime.combine(value, datetime.datetime.min.time())
    return _mongo_compare(op, dtype.prop.place, value)


MONGO_OPERATORS = {
    'lt': '$lt',
    'le': '$lte',
    'gt': '$gt',
    'ge': '$gte',
}


def _mongo_compare(op, column, value):
    if op == 'eq':
        return {column: value}
    else:
        return {column: {MONGO_OPERATORS[op]: value}}


@ufunc.resolver(MongoQueryBuilder, Lower, str)
def eq(env, fn, value):
    value = re.escape(value)
    value = re.compile(f'^{value}$', re.IGNORECASE)
    return env.call('eq', fn.dtype, value)


@ufunc.resolver(MongoQueryBuilder, DataType, object)
def ne(env, dtype, value):
    key = dtype.prop.place
    return {
        '$and': [
            {key: {'$ne': value, '$exists': True}},
            {key: {'$ne': None, '$exists': True}},
        ]
    }


@ufunc.resolver(MongoQueryBuilder, PrimaryKey, object)
def ne(env, dtype, value):
    return {
        '$and': [
            {'__id': {'$ne': value, '$exists': True}},
            {'__id': {'$ne': None, '$exists': True}},
        ]
    }


@ufunc.resolver(MongoQueryBuilder, DateTime, str)
def ne(env, dtype, value):
    value = datetime.datetime.fromisoformat(value)
    return env.call('ne', dtype, value)


@ufunc.resolver(MongoQueryBuilder, Date, str)
def ne(env, dtype, value):
    value = datetime.date.fromisoformat(value)
    value = datetime.datetime.combine(value, datetime.datetime.min.time())
    return env.call('ne', dtype, value)


@ufunc.resolver(MongoQueryBuilder, Lower, str)
def ne(env, fn, value):
    key = fn.dtype.prop.place
    value = re.escape(value)
    value = re.compile(f'^{value}$', re.IGNORECASE)
    return {
        '$and': [
            {key: {'$not': value, '$exists': True}},
            {key: {'$ne': None, '$exists': True}},
        ]
    }


def _ensure_non_empty(op, s):
    if s == '':
        raise EmptyStringSearch(op=op)


@ufunc.resolver(MongoQueryBuilder, String, str)
def contains(env, dtype, value):
    _ensure_non_empty('contains', value)
    value = re.escape(value)
    value = re.compile(value)
    return {dtype.prop.place: value}


@ufunc.resolver(MongoQueryBuilder, Array, (object, type(None)), names=[
    'eq', 'ne', 'lt', 'le', 'gt', 'ge', 'contains', 'startswith',
])
def compare(env, op, dtype, value):
    return env.call(op, dtype.items.dtype, value)


@ufunc.resolver(MongoQueryBuilder, Lower, str)
def contains(env, fn, value):
    _ensure_non_empty('contains', value)
    value = re.escape(value)
    value = re.compile(value, re.IGNORECASE)
    return {fn.dtype.prop.place: value}


@ufunc.resolver(MongoQueryBuilder, PrimaryKey, str)
def contains(env, dtype, value):
    _ensure_non_empty('contains', value)
    value = re.escape(value)
    value = re.compile(value)
    return {'__id': value}


@ufunc.resolver(MongoQueryBuilder, String, str)
def startswith(env, dtype, value):
    _ensure_non_empty('startswith', value)
    value = re.escape(value)
    value = re.compile('^' + value)
    return {dtype.prop.place: value}


@ufunc.resolver(MongoQueryBuilder, Lower, str)
def startswith(env, fn, value):
    _ensure_non_empty('startswith', value)
    value = re.escape(value)
    value = re.compile('^' + value, re.IGNORECASE)
    return {fn.dtype.prop.place: value}


@ufunc.resolver(MongoQueryBuilder, PrimaryKey, str)
def startswith(env, dtype, value):
    _ensure_non_empty('startswith', value)
    value = re.escape(value)
    value = re.compile('^' + value)
    return {'__id': value}


FUNCS = [
    'lower',
    'upper',
    'recurse',
]


@ufunc.resolver(MongoQueryBuilder, Bind, names=FUNCS)
def func(env, name, field):
    prop = env.resolve_property(field)
    return env.call(name, prop.dtype)


@ufunc.resolver(MongoQueryBuilder, String)
def lower(env, dtype):
    return Lower(dtype)


@ufunc.resolver(MongoQueryBuilder, Recurse)
def lower(env, recurse):
    return Recurse([env.call('lower', arg) for arg in recurse.args])


@ufunc.resolver(MongoQueryBuilder, Bind)
def recurse(env, field):
    if field.name in env.model.leafprops:
        return Recurse([prop.dtype for prop in env.model.leafprops[field.name]])
    else:
        raise exceptions.FieldNotInResource(env.model, property=field.name)


@ufunc.resolver(MongoQueryBuilder, Recurse, object, names=[
    'eq', 'ne', 'lt', 'le', 'gt', 'ge', 'contains', 'startswith',
])
def recurse(env, op, recurse, value):
    return env.call('or', [
        env.call(op, arg, value)
        for arg in recurse.args
    ])


@ufunc.resolver(MongoQueryBuilder, Expr, name='any')
def any_(env, expr):
    args, kwargs = expr.resolve(env)
    op, field, *args = args
    if isinstance(op, Bind):
        op = op.name
    return env.call('or', [
        env.call(op, field, arg)
        for arg in args
    ])


@ufunc.resolver(MongoQueryBuilder, Expr)
def sort(env, expr):
    args, kwargs = expr.resolve(env)
    env.sort = [
        env.call('sort', arg) for arg in args
    ]


@ufunc.resolver(MongoQueryBuilder, Bind)
def sort(env, field):
    prop = env.resolve_property(field)
    return env.call('asc', prop.dtype)


@ufunc.resolver(MongoQueryBuilder, Positive)
def sort(env, sign):
    return env.call('asc', sign.arg)


@ufunc.resolver(MongoQueryBuilder, Negative)
def sort(env, sign):
    return env.call('desc', sign.arg)


@ufunc.resolver(MongoQueryBuilder, DataType)
def asc(env, dtype):
    return dtype.prop.place, pymongo.ASCENDING


@ufunc.resolver(MongoQueryBuilder, DataType)
def desc(env, dtype):
    return dtype.prop.place, pymongo.DESCENDING


@ufunc.resolver(MongoQueryBuilder, PrimaryKey)
def asc(env, dtype):
    return '__id', pymongo.ASCENDING


@ufunc.resolver(MongoQueryBuilder, PrimaryKey)
def desc(env, dtype):
    return '__id', pymongo.DESCENDING


@ufunc.resolver(MongoQueryBuilder, Bind)
def negative(env, field: Bind) -> Negative:
    prop = env.resolve_property(field)
    return Negative(prop.dtype)


@ufunc.resolver(MongoQueryBuilder, Bind)
def positive(env, field: Bind) -> Positive:
    prop = env.resolve_property(field)
    return Positive(prop.dtype)


@ufunc.resolver(MongoQueryBuilder, DataType)
def negative(env, dtype) -> Negative:
    return Negative(dtype)


@ufunc.resolver(MongoQueryBuilder, DataType)
def positive(env, dtype) -> Positive:
    return Positive(dtype)
